import math
import sys
from copy import deepcopy

import numpy as np
import torch
from torchvision.ops import nms as torchvision_nms

from ...utils.convenience import to_numpy


def clip_bboxes(bboxes, height, width):
    """[summary]
    :param bboxes: x, y, w, h, or x1, y1, x2, y2
    :type bboxes: [type]
    """
    bboxes[:, 0] = np.clip(bboxes[:, 0], 0, width - sys.float_info.epsilon)
    bboxes[:, 1] = np.clip(bboxes[:, 1], 0, height - sys.float_info.epsilon)
    bboxes[:, 2] = np.clip(bboxes[:, 2], 0, width - sys.float_info.epsilon)
    bboxes[:, 3] = np.clip(bboxes[:, 3], 0, height - sys.float_info.epsilon)
    return bboxes


def denormalize_bboxes_for_image(bboxes, height, width):
    if len(bboxes) == 0:
        return bboxes

    bboxes = deepcopy(bboxes)
    bboxes[:, 0] *= width
    bboxes[:, 1] *= height
    bboxes[:, 2] *= width
    bboxes[:, 3] *= height

    return bboxes


def normalize_bboxes_for_image(bboxes, height, width):
    if len(bboxes) == 0:
        return bboxes

    bboxes = deepcopy(bboxes)
    bboxes[:, 0] /= width
    bboxes[:, 1] /= height
    bboxes[:, 2] /= width
    bboxes[:, 3] /= height
    return bboxes


def xyxy_to_cxcywh(x):
    if len(x) == 0:
        return x

    # Convert nx4 boxes from [x1, y1, x2, y2] to [center_x, center_y, w, h] where xy1=top-left, xy2=bottom-right
    y = torch.zeros_like(x) if isinstance(x, torch.Tensor) else np.zeros_like(x)
    y[:, 0] = (x[:, 0] + x[:, 2]) / 2  # x center
    y[:, 1] = (x[:, 1] + x[:, 3]) / 2  # y center
    y[:, 2] = x[:, 2] - x[:, 0]  # width
    y[:, 3] = x[:, 3] - x[:, 1]  # height
    return y


def cxcywh_to_xyxy(x):
    if len(x) == 0:
        return x

    # Convert nx4 boxes from [center_x, center_y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = torch.zeros_like(x) if isinstance(x, torch.Tensor) else np.zeros_like(x)
    y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
    y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
    y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
    return y


def xywh_to_xyxy(x):
    if len(x) == 0:
        return x

    # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2] where xy1=top-left, xy2=bottom-right
    y = torch.zeros_like(x) if isinstance(x, torch.Tensor) else np.zeros_like(x)
    y[:, 0] = x[:, 0]  # top left x
    y[:, 1] = x[:, 1]  # top left y
    y[:, 2] = x[:, 0] + x[:, 2]  # bottom right x
    y[:, 3] = x[:, 1] + x[:, 3]  # bottom right y
    return y


def bboxes_to_tensor(bboxes, scale_fn=None, format_fn=None):
    bboxes = deepcopy(bboxes)

    if scale_fn is not None:
        bboxes = scale_fn(bboxes)

    if format_fn is not None:
        bboxes = format_fn(bboxes)

    bboxes = torch.from_numpy(bboxes)

    return bboxes


def tensor_to_bboxes(tensor, unscale_fn=None, format_fn=None):
    bboxes = to_numpy(tensor)

    if format_fn is not None:
        bboxes = format_fn(bboxes)

    if unscale_fn is not None:
        bboxes = unscale_fn(bboxes)

    return bboxes


def iou_xywh_numpy(boxes1, boxes2):
    """
    :param boxes1: boxes1和boxes2的shape可以不相同，但是需要满足广播机制
    :param boxes2: 且需要保证最后一维为坐标维，以及坐标的存储结构为(x,y,w,h)，其中(x,y)是bbox的中心坐标
    :return: 返回boxes1和boxes2的IOU，IOU的shape为boxes1和boxes2广播后的shape[:-1]
    """
    boxes1 = np.array(boxes1)
    boxes2 = np.array(boxes2)

    boxes1_area = boxes1[..., 2] * boxes1[..., 3]
    boxes2_area = boxes2[..., 2] * boxes2[..., 3]

    # 分别计算出boxes1和boxes2的左上角坐标、右下角坐标
    # 存储结构为(xmin, ymin, xmax, ymax)，其中(xmin,ymin)是bbox的左上角坐标，(xmax,ymax)是bbox的右下角坐标
    boxes1 = np.concatenate(
        [
            boxes1[..., :2] - boxes1[..., 2:] * 0.5,
            boxes1[..., :2] + boxes1[..., 2:] * 0.5,
        ],
        axis=-1,
    )
    boxes2 = np.concatenate(
        [
            boxes2[..., :2] - boxes2[..., 2:] * 0.5,
            boxes2[..., :2] + boxes2[..., 2:] * 0.5,
        ],
        axis=-1,
    )

    # 计算出boxes1与boxes1相交部分的左上角坐标、右下角坐标
    left_up = np.maximum(boxes1[..., :2], boxes2[..., :2])
    right_down = np.minimum(boxes1[..., 2:], boxes2[..., 2:])

    # 因为两个boxes没有交集时，(right_down - left_up) < 0，所以maximum可以保证当两个boxes没有交集时，它们之间的iou为0
    inter_section = np.maximum(right_down - left_up, 0.0)
    inter_area = inter_section[..., 0] * inter_section[..., 1]
    union_area = boxes1_area + boxes2_area - inter_area
    IOU = 1.0 * inter_area / union_area
    return IOU


def ciou_xywh_torch(boxes1, boxes2):
    """
    cal CIOU of two boxes or batch boxes
    :param boxes1:[xmin,ymin,xmax,ymax] or
                [[xmin,ymin,xmax,ymax],[xmin,ymin,xmax,ymax],...]
    :param boxes2:[xmin,ymin,xmax,ymax]
    :return:
    """
    # xywh->xyxy
    boxes1 = torch.cat(
        [
            boxes1[..., :2] - boxes1[..., 2:] * 0.5,
            boxes1[..., :2] + boxes1[..., 2:] * 0.5,
        ],
        dim=-1,
    )
    boxes2 = torch.cat(
        [
            boxes2[..., :2] - boxes2[..., 2:] * 0.5,
            boxes2[..., :2] + boxes2[..., 2:] * 0.5,
        ],
        dim=-1,
    )

    boxes1 = torch.cat(
        [
            torch.min(boxes1[..., :2], boxes1[..., 2:]),
            torch.max(boxes1[..., :2], boxes1[..., 2:]),
        ],
        dim=-1,
    )
    boxes2 = torch.cat(
        [
            torch.min(boxes2[..., :2], boxes2[..., 2:]),
            torch.max(boxes2[..., :2], boxes2[..., 2:]),
        ],
        dim=-1,
    )

    boxes1_area = (boxes1[..., 2] - boxes1[..., 0]) * (boxes1[..., 3] - boxes1[..., 1])
    boxes2_area = (boxes2[..., 2] - boxes2[..., 0]) * (boxes2[..., 3] - boxes2[..., 1])

    inter_left_up = torch.max(boxes1[..., :2], boxes2[..., :2])
    inter_right_down = torch.min(boxes1[..., 2:], boxes2[..., 2:])
    inter_section = torch.max(
        inter_right_down - inter_left_up, torch.zeros_like(inter_right_down)
    )
    inter_area = inter_section[..., 0] * inter_section[..., 1]
    union_area = boxes1_area + boxes2_area - inter_area
    ious = 1.0 * inter_area / union_area

    # cal outer boxes
    outer_left_up = torch.min(boxes1[..., :2], boxes2[..., :2])
    outer_right_down = torch.max(boxes1[..., 2:], boxes2[..., 2:])
    outer = torch.max(
        outer_right_down - outer_left_up, torch.zeros_like(inter_right_down)
    )
    outer_diagonal_line = torch.pow(outer[..., 0], 2) + torch.pow(outer[..., 1], 2)

    # cal center distance
    boxes1_center = (boxes1[..., :2] + boxes1[..., 2:]) * 0.5
    boxes2_center = (boxes2[..., :2] + boxes2[..., 2:]) * 0.5
    center_dis = torch.pow(
        boxes1_center[..., 0] - boxes2_center[..., 0], 2
    ) + torch.pow(boxes1_center[..., 1] - boxes2_center[..., 1], 2)

    # cal penalty term
    # cal width,height
    boxes1_size = torch.max(
        boxes1[..., 2:] - boxes1[..., :2], torch.zeros_like(inter_right_down)
    )
    boxes2_size = torch.max(
        boxes2[..., 2:] - boxes2[..., :2], torch.zeros_like(inter_right_down)
    )
    v = (4 / (math.pi ** 2)) * torch.pow(
        torch.atan((boxes1_size[..., 0] / torch.clamp(boxes1_size[..., 1], min=1e-6)))
        - torch.atan(
            (boxes2_size[..., 0] / torch.clamp(boxes2_size[..., 1], min=1e-6))
        ),
        2,
    )
    alpha = v / (1 - ious + v)

    # cal ciou
    cious = ious - (center_dis / outer_diagonal_line + alpha * v)

    return cious


def convert_bbox_yolo_to_pascal(bboxes, image):
    height, width, _ = image.shape
    return cxcywh_to_xyxy(
        denormalize_bboxes_for_image(bboxes, height=height, width=width)
    )


def nms_with_scores(boxes, scores, labels, iou_threshold=0.1):
    boxes = torch.from_numpy(boxes) if isinstance(boxes, np.ndarray) else boxes
    scores = torch.from_numpy(scores) if isinstance(scores, np.ndarray) else scores
    labels = torch.from_numpy(labels) if isinstance(labels, np.ndarray) else labels
    nms_idxs = torchvision_nms(boxes, scores, iou_threshold)
    return {
        "boxes": boxes[nms_idxs],
        "scores": scores[nms_idxs],
        "labels": labels[nms_idxs],
    }


lookup = {
    "bboxes_to_tensor": bboxes_to_tensor,
    "ciou_xywh_torch": ciou_xywh_torch,
    "clip_bboxes": clip_bboxes,
    "convert_bbox_yolo_to_pascal": convert_bbox_yolo_to_pascal,
    "denormalize_bboxes_for_image": denormalize_bboxes_for_image,
    "iou_xywh_numpy": iou_xywh_numpy,
    "normalize_bboxes_for_image": normalize_bboxes_for_image,
    "xyxy_to_cxcywh": xyxy_to_cxcywh,
    "tensor_to_bboxes": tensor_to_bboxes,
    "cxcywh_to_xyxy": cxcywh_to_xyxy,
    "cxcywh_to_xyxy": cxcywh_to_xyxy,
    "nms_with_scores": nms_with_scores,
}



import logging
import os
import re
import time
from enum import Enum

import numpy as np
import torch

from .convenience import if_none, is_listy


def num_cpus() -> int:
    try:
        return len(os.sched_getaffinity(0))
    except AttributeError:
        return os.cpu_count()


default_cpus = min(1, num_cpus())
default_device = (
    torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
)


# https://stackoverflow.com/questions/29831489/convert-array-of-indices-to-1-hot-encoded-numpy-array
def make_one_hot(x, n_classes=None, channel_first=False, dtype=np.float32):
    """1-hot encode x with the max value n (computed from data if n is None)."""
    x = np.asarray(x)
    n = np.max(x) + 1 if n_classes is None else n_classes
    one_hot = np.eye(n, dtype=dtype)[x]
    if not channel_first:
        return one_hot
    return one_hot.transpose(2, 0, 1)


def tensorify(lst):
    """
    List must be nested list of tensors (with no varying lengths within a dimension).
    Nested list of nested lengths [D1, D2, ... DN] -> tensor([D1, D2, ..., DN)

    :return: nested list D
    """
    # base case, if the current list is not nested anymore, make it into tensor
    if type(lst[0]) != list:
        if type(lst) == torch.Tensor:
            return lst
        elif type(lst[0]) == torch.Tensor:
            return torch.stack(lst, dim=0)
        else:  # if the elements of lst are floats or something like that
            return torch.tensor(lst)
    current_dimension_i = len(lst)
    for d_i in range(current_dimension_i):
        tensor = tensorify(lst[d_i])
        lst[d_i] = tensor
    # end of loop lst[d_i] = tensor([D_i, ... D_0])
    tensor_lst = torch.stack(lst, dim=0)
    return tensor_lst


def get_device_for_tensor(tensor):
    return torch.device(f"cuda:{tensor.get_device()}")


def to_device(tensors, device):
    device = if_none(device, default_device)
    if is_listy(tensors):
        return [to_device(o, device) for o in tensors]
    return tensors.to(device)


lookup = {
    "num_cpus": num_cpus,
    "default_cpus": default_cpus,
    "default_device": default_device,
    "make_one_hot": make_one_hot,
    "get_device_for_tensor": get_device_for_tensor,
    "to_device": to_device,
    "tensorify": tensorify,
}



from .cadene import lookup as cadene_model_lookup
from .torchvision import lookup as torchvision_lookup
from .unet import lookup as unet_lookup

models_lookup = {**cadene_model_lookup, **unet_lookup, **torchvision_lookup}



from textwrap import dedent

import yaml


def create_seldon_deployment_manifest(
    image,
    deployment_name="app",
    predictor_name="default",
    spec_name="classifier",
    container_name="classifier",
    gateway_name=None,
    namespace="seldon",
    replicas=1,
):
    manifest = yaml.safe_load(
        dedent(
            f"""
        apiVersion: machinelearning.seldon.io/v1
        kind: SeldonDeployment
        metadata:
          name: {deployment_name}
          namespace: {namespace}
        spec:
          name: {spec_name}
          protocol: seldon
          predictors:
            - name: {predictor_name}
              graph:
                name: classifier
              componentSpecs:
                - spec:
                    containers:
                      - image: {image}
                        name: {container_name}
                        resources:
                          limits:
                            nvidia.com/gpu: 1
                        livenessProbe:
                          failureThreshold: 3
                          initialDelaySeconds: 60
                          periodSeconds: 5
                          successThreshold: 1
                          httpGet:
                            path: /health/status
                            port: http
                            scheme: HTTP
                          timeoutSeconds: 1
                        readinessProbe:
                          failureThreshold: 3
                          initialDelaySeconds: 20
                          periodSeconds: 5
                          successThreshold: 1
                          httpGet:
                              path: /health/status
                              port: http
                          timeoutSeconds: 1
              replicas: {replicas}
    """
        )
    )

    if gateway_name is not None:
        manifest["spec"].get("annotations", {})[
            "seldon.io/istio-gateway"
        ] = gateway_name

    return manifest



# vendored in from https://github.dev/scnerd/miniutils

import itertools
import multiprocessing as mp

try:
    from nose.plugins.multiprocess import TimedOutException
except ImportError:
    TimedOutException = TimeoutError
import random
import warnings

try:
    from tqdm import tqdm as _tqdm

    try:  # pragma: nocover
        # Check if we're in a Jupyter notebook... if so, use the ipywidgets progress bar instead
        from IPython import get_ipython

        if type(get_ipython()).__module__.startswith("ipykernel."):
            from tqdm import tqdm_notebook as _tqdm
    except (ImportError, NameError):  # pragma: nocover
        # IPython isn't even installed, or we're not in it
        pass
except ImportError:  # pragma: nocover
    # noinspection PyUnusedLocal
    def _tqdm(iterable, *a, **kw):
        return iterable


def progbar(iterable, *a, verbose=True, **kw):
    """Prints a progress bar as the iterable is iterated over

    :param iterable: The iterator to iterate over
    :param a: Arguments to get passed to tqdm (or tqdm_notebook, if in a Jupyter notebook)
    :param verbose: Whether or not to print the progress bar at all
    :param kw: Keyword arguments to get passed to tqdm
    :return: The iterable that will report a progress bar
    """
    iterable = range(iterable) if isinstance(iterable, int) else iterable
    if verbose:
        return _tqdm(iterable, *a, **kw)
    else:
        return iterable


def _fun(f, q_in, q_out, flatten, star):  # pragma: no cover
    try:
        while True:
            i, x = q_in.get()
            if i is None:
                break
            out = f(*x) if star else f(x)
            if flatten:
                for j, o in enumerate(out):
                    q_out.put(((i, j), o))
                q_out.put((None, None))
            else:
                q_out.put((i, out))
    except BaseException as ex:
        q_out.put((None, ex))


def _parallel_progbar_launch(
    mapper,
    iterable,
    nprocs=None,
    starmap=False,
    flatmap=False,
    shuffle=False,
    verbose=True,
    verbose_flatmap=None,
    max_cache=-1,
    timeout=1,
    **kwargs
):

    # Shuffle the iterable if requested, to make the parallel execution potentially more uniform in runtime
    enumerated_iterable = enumerate(iterable)
    if shuffle:
        enumerated_iterable = list(enumerated_iterable)
        # ids = [i for i in sorted(range(len(iterable)), key=lambda x: random.random())]
        # iterable = (iterable[i] for i in ids)
        random.shuffle(
            enumerated_iterable
        )  # Is this going to be expensive for large lists of large objects?

    # Check that we don't launch more processes than there are elements to map (if that's knowable)
    nprocs = nprocs or mp.cpu_count()
    try:
        nprocs = max(1, min(len(iterable), nprocs))
    except TypeError:
        pass

    # Set up multiprocessing management for mapping
    q_in = mp.Queue()
    q_out = mp.Queue(max_cache)

    procs = [
        mp.Process(target=_fun, args=(mapper, q_in, q_out, flatmap, starmap))
        for _ in range(nprocs)
    ]
    for p in procs:
        p.daemon = True
        p.start()

    # Doing it this way prevents us from storing locally an entire list of the input values unnecessarily, and still
    # gets us the number of elements sent for processing
    sent = (q_in.put((i, x)) for i, x in enumerated_iterable)
    num_sent = sum(1 for _ in sent)
    for _ in range(nprocs):
        # Send out a flag for each process to terminate once all elements are processed
        q_in.put((None, None))

    # Fetch the mapped results from the output queue, printing a progress bar as you go
    if flatmap:
        # If we're flat mapping, then we'll keep separate progress of all returned results (an unknown number) and how
        # many inputs are complete (a known number). The outer loop will track the latter, and the inner loop the former
        results = (
            q_out.get()
            for _ in progbar(
                itertools.count(),
                verbose=verbose if verbose_flatmap is None else verbose_flatmap,
                **kwargs
            )
        )
        for _ in progbar(num_sent, verbose=verbose):
            for i, x in results:
                # When we're flagged that an input is done being returned in the queue, break the inner loop to make
                # the "completed inputs" progress bar tick
                if i is None:
                    if x is None:
                        break
                    else:
                        raise x
                yield i, x
    else:
        for i, x in (q_out.get() for _ in progbar(num_sent, verbose=verbose, **kwargs)):
            if i is None:
                raise x
            yield i, x

    # Clean up
    for p in procs:
        try:
            p.join(timeout)
        except (TimeoutError, mp.TimeoutError, TimedOutException):  # pragma: nocover
            warnings.warn(
                "parallel_progbar mapping process failed to close properly (check error output)"
            )


def parallel_progbar(*args, **kwargs):
    """Performs a parallel mapping of the given iterable, reporting a progress bar as values get returned

    :param mapper: The mapping function to apply to elements of the iterable
    :param iterable: The iterable to map
    :param nprocs: The number of processes (defaults to the number of cpu's)
    :param starmap: If true, the iterable is expected to contain tuples and the mapper function gets each element of a
        tuple as an argument
    :param flatmap: If true, flatten out the returned values if the mapper function returns a list of objects
    :param shuffle: If true, randomly sort the elements before processing them. This might help provide more uniform
        runtimes if processing different objects takes different amounts of time.
    :param verbose: Whether or not to print the progress bar
    :param verbose_flatmap: If performing a flatmap, whether or not to report each object as it's returned
    :param timeout: The number of seconds to wait for each worker process after completing
    :param kwargs: Any other keyword arguments to pass to the progress bar (see ``progbar``)
    :return: A list of the returned objects, in the same order as provided
    """

    results = _parallel_progbar_launch(*args, **kwargs)
    return [x for i, x in sorted(results, key=lambda p: p[0])]


def iparallel_progbar(*args, **kwargs):
    """Performs a parallel mapping of the given iterable, reporting a progress bar as values get returned. Yields
    objects as soon as they're computed, but does not guarantee that they'll be in the correct order.

    :param mapper: The mapping function to apply to elements of the iterable
    :param iterable: The iterable to map
    :param nprocs: The number of processes (defaults to the number of cpu's)
    :param starmap: If true, the iterable is expected to contain tuples and the mapper function gets each element of a
        tuple as an argument
    :param flatmap: If true, flatten out the returned values if the mapper function returns a list of objects
    :param shuffle: If true, randomly sort the elements before processing them. This might help provide more uniform
        runtimes if processing different objects takes different amounts of time.
    :param verbose: Whether or not to print the progress bar
    :param verbose_flatmap: If performing a flatmap, whether or not to report each object as it's returned
    :param max_cache: Maximum number of mapped objects to permit in the queue at once
    :param timeout: The number of seconds to wait for each worker process after completing
    :param kwargs: Any other keyword arguments to pass to the progress bar (see ``progbar``)
    :return: A list of the returned objects, in whatever order they're done being computed
    """

    results = _parallel_progbar_launch(*args, **kwargs)
    return (x for i, x in results)



from torch import Tensor

from .convenience import is_listy


class Hooks:
    "Create several hooks on the modules in `ms` with `hook_func`."

    def __init__(self, modules, hook_fn, is_forward=True, detach=True):
        self.hooks = [Hook(m, hook_fn, is_forward, detach) for m in modules]

    def __getitem__(self, i):
        return self.hooks[i]

    def __len__(self):
        return len(self.hooks)

    def __iter__(self):
        return iter(self.hooks)

    @property
    def stored(self):
        return [o.stored for o in self]

    def remove(self):
        "Remove the hooks from the model."
        for h in self.hooks:
            h.remove()

    def __enter__(self, *args):
        return self

    def __exit__(self, *args):
        self.remove()


class Hook:
    "Create a hook on `m` with `hook_func`."

    def __init__(self, m, hook_fn, is_forward=True, detach=True):
        self.hook_fn, self.detach, self.stored = hook_fn, detach, None
        register_hook_fn = (
            m.register_forward_hook if is_forward else m.register_backward_hook
        )
        self.hook = register_hook_fn(self.hook_fn_wrapper)
        self.removed = False

    def hook_fn_wrapper(self, module, input, output):
        "Applies `hook_func` to `module`, `input`, `output`."
        if self.detach:
            input = (o.detach() for o in input) if is_listy(input) else input.detach()
            output = (
                (o.detach() for o in output) if is_listy(output) else output.detach()
            )
        self.stored = self.hook_fn(module, input, output)

    def remove(self):
        "Remove the hook from the model."
        if not self.removed:
            self.hook.remove()
            self.removed = True

    def __enter__(self, *args):
        return self

    def __exit__(self, *args):
        self.remove()


def hook_output(module, detach=True, grad=False):
    "Return a `Hook` that stores activations of `module` in `self.stored`"
    return Hook(module, _hook_inner, detach=detach, is_forward=not grad)


def hook_outputs(modules, detach=True, grad=False):
    "Return `Hooks` that store activations of all `modules` in `self.stored`"
    return Hooks(modules, _hook_inner, detach=detach, is_forward=not grad)


def _hook_inner(m, i, o):
    return o if isinstance(o, Tensor) else o if is_listy(o) else list(o)


lookup = {
    "Hooks": Hooks,
    "Hook": Hook,
    "hook_output": hook_output,
    "hook_outputs": hook_outputs,
}



try:
    from .activeloop import *
    from .activeloop import lookup as activeloop_lookup
    from .label_per_folder import *
    from .label_per_folder import lookup as label_per_folder_lookup

    lookup = {**label_per_folder_lookup, **activeloop_lookup}
except Exception:
    from .label_per_folder import *
    from .label_per_folder import lookup as label_per_folder_lookup

    lookup = {**label_per_folder_lookup}



import docker


def run_mlflow_model_image(image, port="5000", gpus=None):
    client = docker.from_env()
    run_params = {"image": image, "ports": {"5000/tcp": str(port)}, "detach": True}
    if gpus is not None:
        run_params["runtime"] = "nvidia"
    container = client.containers.run(**run_params)
    return container


def run_seldon_model_image(image, rest_port="9000", grpc_port="5000", gpus=None):
    client = docker.from_env()
    run_params = {
        "image": image,
        "ports": {"9000/tcp": str(rest_port), "5000/tcp": str(grpc_port)},
        "detach": True,
    }
    if gpus is not None:
        run_params["runtime"] = "nvidia"
    container = client.containers.run(**run_params)
    return container



import numpy as np
import torch
import torch.utils

from .....utils.convenience import compose


class ImageBoundingBoxDetectionAugmentationMixin:
    def augment(self, image, bboxes, labels):
        if self.augment_fn is None:
            return image, bboxes, labels

        result = self.augment_fn(image=image, bboxes=bboxes, class_labels=labels)
        image = result["image"]
        bboxes = np.array(result["bboxes"])
        labels = np.array(result["class_labels"])
        return image, bboxes, labels


class ImageBoundingBoxDetectionDatasetV2(
    torch.utils.data.Dataset, ImageBoundingBoxDetectionAugmentationMixin
):
    def __init__(
        self,
        samples,
        transform_sample_fns=[lambda x: x],
        augment_fn=None,
        tile_fn=None,
        convert_image_to_tensor_fn=lambda image: torch.from_numpy(image),
        convert_bboxes_to_tensor_fn=lambda bboxes: torch.from_numpy(bboxes),
    ):
        self.samples = samples
        self.transform_sample_fns = compose(transform_sample_fns)
        self.augment_fn = augment_fn
        self.tile_fn = tile_fn
        self.convert_image_to_tensor_fn = convert_image_to_tensor_fn
        self.convert_bboxes_to_tensor_fn = convert_bboxes_to_tensor_fn

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, i):
        sample = self.transform_sample_fns(self.samples[i])
        data = {}
        data["name"] = sample.name
        image, bboxes, labels, _ = sample.data
        image, bboxes, labels = self.augment(image, bboxes, labels)

        if self.tile_fn is not None:
            image_tiles = self.tile_fn(image)

        if self.convert_image_to_tensor_fn is not None:
            if self.tile_fn is not None:
                data["image"] = torch.stack(
                    [
                        self.convert_image_to_tensor_fn(image_tile)
                        for image_tile in image_tiles
                    ]
                )
            else:
                data["image"] = self.convert_image_to_tensor_fn(image)

        if sample.labels_exists:
            if self.tile_fn is not None:
                bboxes_tiles = self.tile_fn(bboxes)

            if self.convert_bboxes_to_tensor_fn is not None:
                if self.tile_fn is not None:
                    data["boxes"] = torch.stack(
                        [
                            self.convert_bboxes_to_tensor_fn(bbox_tile)
                            for bbox_tile in bboxes_tiles
                        ]
                    )
                else:
                    data["boxes"] = self.convert_bboxes_to_tensor_fn(bboxes)
            data["labels"] = torch.from_numpy(labels)
        return data


def image_bounding_box_detection_torchvision_collate_fn(samples):
    # batch = {}
    # names = [s.name for s in samples]
    # batch["names"] = names
    # batch["images"] = [s.image.tensor.float() for s in samples]

    # has_bbox_and_labels = samples[0].labels_exists and samples[0].bboxes_exists
    # if has_bbox_and_labels:
    #     batch["targets"] = []
    #     for s in samples:
    #         boxes = s.bboxes.tensor.float()
    #         labels = s.labels.tensor.type(torch.int64)
    #         if len(boxes) == 0:
    #             boxes = torch.zeros((0, 4), dtype=torch.float32)
    #             labels = torch.zeros((1, 1), dtype=torch.int64)
    #         batch["targets"].append({"boxes": boxes, "labels": labels})
    batch = {}
    batch["names"] = [s["name"] for s in samples]
    batch["images"] = [s["image"].float() for s in samples]

    has_bbox_and_labels = "labels" in samples[0] and "boxes" in samples[0]
    if has_bbox_and_labels:
        batch["targets"] = []
        for sample in samples:
            boxes = sample["boxes"].float()
            labels = sample["labels"].type(torch.int64)
            if len(boxes) == 0:
                boxes = torch.zeros((0, 4), dtype=torch.float32)
                labels = torch.zeros((1, 1), dtype=torch.int64)
            batch["targets"].append({"boxes": boxes, "labels": labels})
    return batch


lookup = {
    # "ImageBoundingBoxDetectionDataset": ImageBoundingBoxDetectionDataset,
    "ImageBoundingBoxDetectionDatasetV2": ImageBoundingBoxDetectionDatasetV2,
    "image_bounding_box_detection_torchvision_collate_fn": image_bounding_box_detection_torchvision_collate_fn,
}



from functools import partial

import numpy as np
import torch

from ...utils.training import get_device_for_tensor, to_device


def normalize_image(image, *, mean, std):
    assert image.dtype == np.float32
    denominator = np.reciprocal(std, dtype=np.float32)
    image -= mean
    image *= denominator
    return image


def unnormalize_image(image, *, mean, std):
    assert image.dtype == np.float32
    denominator = np.reciprocal(std, dtype=np.float32)
    image /= denominator
    image += mean
    return image


def normalize_tensor(tensor, *, mean, std):
    mean = torch.Tensor(mean)
    std = torch.Tensor(std)

    if tensor.is_cuda:
        device = get_device_for_tensor(tensor)
        mean = to_device(mean, device)
        std = to_device(std, device)

    tensor = (tensor - mean[..., None, None]) / std[..., None, None]
    return tensor


def unnormalize_tensor(tensor, *, mean, std):
    mean = torch.Tensor(mean)
    std = torch.Tensor(std)

    if tensor.is_cuda:
        device = get_device_for_tensor(tensor)
        mean = to_device(mean, device)
        std = to_device(std, device)

    tensor = tensor * std[..., None, None] + mean[..., None, None]
    return tensor


IMAGE_NET_STATS = {"mean": [0.485, 0.456, 0.406], "std": [0.229, 0.224, 0.225]}

FOUR_CHANNEL_PNASNET5LARGE_STATS = {
    "mean": [0.5, 0.5, 0.5, 0.5],
    "std": [0.5, 0.5, 0.5, 0.5],
}

FOUR_CHANNEL_IMAGE_NET_STATS = {
    "mean": [0.485, 0.456, 0.406, 0.485],
    "std": [0.229, 0.224, 0.224, 0.229],
}
image_net_normalize = partial(normalize_tensor, **IMAGE_NET_STATS)
image_net_denormalize = partial(unnormalize_tensor, **IMAGE_NET_STATS)

four_channel_image_net_normalize = partial(
    normalize_tensor, **FOUR_CHANNEL_IMAGE_NET_STATS
)
four_channel_image_net_denormalize = partial(
    unnormalize_tensor, **FOUR_CHANNEL_IMAGE_NET_STATS
)

four_channel_pnasnet5large_normalize = partial(
    normalize_tensor, **FOUR_CHANNEL_PNASNET5LARGE_STATS
)
four_channel_pnasnet5large_denormalize = partial(
    unnormalize_tensor, **FOUR_CHANNEL_PNASNET5LARGE_STATS
)

lookup = {
    "normalize_tensor": normalize_tensor,
    "unnormalize_tensor": unnormalize_tensor,
    "normalize_image": normalize_image,
    "unnormalize_image": unnormalize_image,
    "image_net_denormalize": image_net_denormalize,
    "four_channel_image_net_unnormalize": four_channel_image_net_denormalize,
    "four_channel_pnasnet5large_unnormalize": four_channel_image_net_denormalize,
    "image_net_normalize": image_net_normalize,
    "four_channel_image_net_normalize": four_channel_image_net_normalize,
    "four_channel_pnasnet5large_normalize": four_channel_image_net_normalize,
    "identity": lambda x: x,
}



from tempfile import TemporaryDirectory


def return_boolean(boolean):
    return boolean


def return_negated_boolean(boolean):
    return not boolean


# def partially_construct(fn, **kwargs):
#     return partial(fn, **kwargs)


# def turn_first_arg_into_kwargs(fn):
#     def inner_fn(*args, **kwargs):
#         return fn(fn=kwargs["fn"], kwargs=args[0])

#     return inner_fn


# @turn_first_arg_into_kwargs
# def spread_arguments_before_call(fn, kwargs):
#     return fn(**kwargs)


# def function_composer(fns):
#     def wrapper(arg):
#         for fn in fns:
#             arg = fn(arg)
#         return arg

#     return wrapper


def create_temp_path():
    return TemporaryDirectory().name


lookup = {
    "return_boolean": return_boolean,
    "return_negated_boolean": return_negated_boolean,
    # "partially_construct": partially_construct,
    # "spread_arguments_before_call": spread_arguments_before_call,
    # "function_composer": function_composer,
    "create_temp_path": create_temp_path,
}



from pathlib import Path

import numpy as np
import torch
from tqdm import tqdm

from ..vendored.ngp_pl.datasets.color_utils import read_image
from ..vendored.ngp_pl.datasets.nsvf import NSVFDataset as ngp_pl_NSVFDataset
from ..vendored.ngp_pl.datasets.ray_utils import get_ray_directions


class NSVFDataset(ngp_pl_NSVFDataset):
    @staticmethod
    def create_shift_and_scale_from_bbox_path(bbox_path):
        xyz_min, xyz_max = np.loadtxt(str(bbox_path))[:6].reshape(2, 3)
        shift = (xyz_max + xyz_min) / 2
        scale = (xyz_max - xyz_min).max() / 2 * 1.05  # enlarge a little
        return shift, scale

    @staticmethod
    def create_rays_and_poses_from_paths(
        image_paths, pose_paths, shift, scale, image_width, image_height, downsample
    ):
        rays = []
        poses = []
        for img_path, pose in tqdm(zip(image_paths, pose_paths)):
            c2w = np.loadtxt(pose)[:3]
            c2w[:, 3] -= shift
            c2w[:, 3] /= 2 * scale  # to bound the scene inside [-0.5, 0.5]
            poses.append(c2w)
            img = read_image(
                img_path,
                (int(image_width * downsample), int(image_height * downsample)),
            )
            rays.append(img)
        rays = torch.FloatTensor(np.stack(rays))  # (N_images, hw, ?)
        poses = torch.FloatTensor(np.stack(poses))  # (N_images, 3, 4)

        return rays, poses

    @staticmethod
    def create_intrinsic_camera_matrix_from_path(
        path, image_width, image_height, downsample
    ):
        with Path(path).open("r") as f:
            fx = fy = float(f.readline().split()[0]) * downsample
        image_width *= downsample
        image_height *= downsample
        K = np.float32([[fx, 0, image_width / 2], [0, fy, image_height / 2], [0, 0, 1]])
        return K

    def __init__(
        self,
        rays,
        poses,
        K,
        image_width,
        image_height,
        batch_size,
        ray_sampling_strategy,
        split,
    ):
        self.rays = rays
        self.poses = poses
        self.K = torch.FloatTensor(K)
        self.directions = get_ray_directions(image_height, image_width, self.K)
        self.img_wh = (image_width, image_height)
        self.batch_size = batch_size
        self.ray_sampling_strategy = ray_sampling_strategy
        self.split = split


lookup = {"NSVFDataset": NSVFDataset}



from abc import ABC, abstractmethod

from sklearn.model_selection import ShuffleSplit as SKLearnShuffleSplit
from sklearn.model_selection import StratifiedShuffleSplit as SKLearnStratifiedShuffleSplit


class Splitter(ABC):
    @abstractmethod
    def split(self):
        pass


class ShuffleSplit(SKLearnShuffleSplit):
    def __init__(
        self, X, n_splits=10, test_size=None, train_size=None, random_state=None
    ):
        self.X = X
        self.splitter = SKLearnShuffleSplit(
            n_splits=n_splits,
            test_size=test_size,
            train_size=train_size,
            random_state=random_state,
        )

    def split(self):
        for (train_idx, val_idx) in self.splitter.split(self.X):
            yield train_idx, val_idx


class StratifiedShuffleSplit(SKLearnStratifiedShuffleSplit):
    def __init__(
        self, X, y, n_splits=10, test_size=None, train_size=None, random_state=None
    ):
        self.X = X
        self.y = y
        self.splitter = SKLearnShuffleSplit(
            n_splits=n_splits,
            test_size=test_size,
            train_size=train_size,
            random_state=random_state,
        )

    def split(self):
        for (train_idx, val_idx) in self.splitter.split(self.X, self.y):
            yield train_idx, val_idx



from .callback_method_is_called_recorder import *
from .image_bounding_box_detection_recorder import *
from .image_classification_recorder import *
from .image_segmentation_recorder import *
from .loss_recorder import *
from .tensorboard_recorder import *

lookup = {
    "CallbackMethodIsCalled": CallbackMethodIsCalledRecorder,
    "CallbackMethodIsCalledRecorder": CallbackMethodIsCalledRecorder,
    "ImageBoundingBoxDetectionResultRecorder": ImageBoundingBoxDetectionResultRecorder,
    "ImageBoundingBoxDetectionInferenceRecorder": ImageBoundingBoxDetectionInferenceRecorder,
    "ImageClassificationResultRecorder": ImageClassificationResultRecorder,
    "ImageClassificationInferenceRecorder": ImageClassificationInferenceRecorder,
    "ImageSegmentationResultRecorder": ImageSegmentationResultRecorder,
    "ImageSegmentationInferenceRecorder": ImageSegmentationInferenceRecorder,
    "PerSampleLossRecorder": PerSampleLossRecorder,
    "TensorboardRecorder": TensorboardRecorder,
}



import random

import numpy as np


def create_black_image_data(
    width, height, n_channels, torch_format=True, type=np.uint8
):
    image = np.stack([np.zeros((height, width))] * n_channels)

    if torch_format:
        image = image.transpose(1, 2, 0)

    return image.astype(type)


def create_random_labels_data(n_labels, n_classes):
    return np.array([random.randint(0, n_classes) for _ in range(n_labels)])


def create_random_yolo_bboxes_data(n_bboxes):
    bboxes = []
    for _ in range(n_bboxes):
        center_x = random.random()
        center_y = random.random()
        width = abs(1 - center_x) * random.random()
        height = abs(1 - center_y) * random.random()
        bboxes.append([center_x, center_y, width, height])
    return np.array(bboxes)


def create_random_pascal_bboxes_data(max_width, max_height, n_bboxes):
    bboxes = []
    for _ in range(n_bboxes):
        x1 = random.random() * max_width
        y1 = random.random() * max_height
        x2 = abs(1 - x1) * random.random() * max_width
        y2 = abs(1 - y1) * random.random() * max_height
        bboxes.append([x1, y1, x2, y2])
    return np.array(bboxes)



import hub
import numpy as np
from hub.schema import BBox, ClassLabel, Image, Text

from .....integrations.activeloop import ActiveLoopHub, compute_activeloop_element
from ..data import create_image_bboxes_detection_sample


class ImageBoundingBoxDetectionActiveLoopHub(ActiveLoopHub):
    def __init__(self, activeloop_dataset):
        super().__init__(activeloop_dataset)

    @staticmethod
    def create_schema_for_rgb(
        id_max_shape,
        image_max_shape,
        bbox_max_shape,
        label_names=None,
        num_classes=None,
    ):
        assert (label_names is None) ^ (num_classes is None)
        if label_names:
            class_label_args = {"names": label_names}
        if num_classes:
            class_label_args = {"num_classes": num_classes}

        return {
            "id": Text(shape=(None,), max_shape=id_max_shape),
            "image": Image(
                shape=(None, None, 3), dtype="uint8", max_shape=image_max_shape
            ),
            "bboxes": BBox(shape=(None, 4), max_shape=bbox_max_shape),
            "labels": ClassLabel(
                shape=(None,), max_shape=(bbox_max_shape[0],), **class_label_args
            ),
        }

    @classmethod
    def create_from_url(cls, url):
        activeloop_dataset = hub.Dataset(url)
        ds = cls(activeloop_dataset)
        return ds

    @classmethod
    def create_from_url_and_checkout(cls, url, address):
        activeloop_dataset = hub.Dataset(url)
        activeloop_dataset.checkout(address)
        ds = cls(activeloop_dataset)
        return ds

    def create_samples(self):
        return [
            create_image_bboxes_detection_sample(**sample_data)
            for sample_data in self.create_samples_data()
        ]

    def create_samples_data(self, bboxes_format="YOLO"):
        arguments = [
            {
                "name": data["id"].compute(),
                "image_path": data["image"],
                "load_image_fn": compute_activeloop_element,
                "bboxes_path": data["bboxes"],
                "load_bboxes_fn": compute_activeloop_element,
                "labels_path": data["labels"],
                "load_labels_fn": compute_activeloop_element,
                "bboxes_format": bboxes_format,
            }
            for data in self.activeloop_dataset
        ]
        return np.array(arguments)

    def create_sample_data_from_index(self, index, bboxes_format="YOLO"):
        data = self.activeloop_dataset[index]
        return {
            "name": data["id"].compute(),
            "image_path": data["image"],
            "load_image_fn": compute_activeloop_element,
            "bboxes_path": data["bboxes"],
            "load_bboxes_fn": compute_activeloop_element,
            "bboxes_format": bboxes_format,
            "labels_path": data["labels"],
            "load_labels_fn": compute_activeloop_element,
        }

    def create_sample_from_index(self, index, bboxes_format="YOLO"):
        return create_image_bboxes_detection_sample(
            **self.create_sample_data_from_index(index, bboxes_format)
        )

    def update_index_with_sample(self, index, sample):
        self.activeloop_dataset["id"][index] = sample.name
        self.activeloop_dataset["image"][index] = sample.image.data
        self.activeloop_dataset["bboxes"][index] = sample.bboxes.data
        self.activeloop_dataset["labels"][index] = sample.labels.data


def create_image_bounding_box_detection_activeloop_data_from_sample(sample):
    return (
        {
            "id": sample.name,
            "image": sample.image.data,
            "labels": sample.labels.data,
            "bboxes": sample.bboxes.data,
        },
    )


def create_image_bounding_box_detection_activeloop_and_checkout(url, address):
    return ImageBoundingBoxDetectionActiveLoopHub.create_from_url_and_checkout(
        url, address
    )


lookup = {
    "ImageBoundingBoxDetectionActiveLoopHub": ImageBoundingBoxDetectionActiveLoopHub,
    "create_image_bounding_box_detection_activeloop_data_from_sample": create_image_bounding_box_detection_activeloop_data_from_sample,
    "create_image_bounding_box_detection_activeloop_and_checkout": create_image_bounding_box_detection_activeloop_and_checkout,
}



import torch.nn as nn


def cadene_change_last_linear(model_creator, out_features):
    model = model_creator()
    dim_feats = model.last_linear.in_features
    model.last_linear = nn.Linear(dim_feats, out_features)
    return model



import json
from glob import glob
from pathlib import Path

import numpy as np
from p_tqdm import p_map

from ..data.sample import create_image_bboxes_detection_sample


def roheboam_sample_data_paths_from_path(root_path, category):
    image_paths = []
    label_paths = []
    for sample_directory in Path(root_path).glob("*"):
        metadata_path = sample_directory / "metadata.json"
        if not (metadata_path).exists():
            continue

        metadata = json.load(open(metadata_path))
        sample_category = metadata["category"]
        imagePath = metadata["imagePath"]
        labelPath = metadata["labelPath"]
        assert metadata["labelType"] == "ULTRALYTICS"

        absolute_image_path = metadata_path.parent / imagePath
        absolute_label_path = (
            metadata_path.parent / labelPath if labelPath is not None else None
        )

        if sample_category == category:
            image_paths.append(absolute_image_path)
            label_paths.append(absolute_label_path)

    return image_paths, label_paths


def load_label_from_path_roheboam_format(path):
    labels = np.loadtxt(path).reshape(-1, 5)

    if len(labels) == 0:
        return np.array([])

    return np.array(labels[:, 0].astype(np.int32))


def load_bboxes_from_path_roheboam_format(path):
    bboxes = np.loadtxt(path).reshape(-1, 5)

    if len(bboxes) == 0:
        return np.array([])

    return np.array(bboxes[:, 1:])


def create_bboxes_detection_samples_data_from_roheboam_format(
    image_paths,
    labels_paths,
    bboxes_format,
    transform_to_bboxes_format=None,
    labels_to_keep=None,
    transform_label_map=None,
):
    return np.array(
        [
            {
                "image_path": image_path,
                "labels_path": labels_path,
                "load_labels_fn": load_label_from_path_roheboam_format,
                "bboxes_path": labels_path,
                "load_bboxes_fn": load_bboxes_from_path_roheboam_format,
                "bboxes_format": bboxes_format,
                "transform_to_bboxes_format": transform_to_bboxes_format,
                "labels_to_keep": labels_to_keep,
                "transform_label_map": transform_label_map,
            }
            for image_path, labels_path in zip(image_paths, labels_paths)
        ]
    )


def create_bboxes_detection_samples_from_roheboam_format(
    image_paths,
    labels_paths,
    bboxes_format,
    transform_to_bboxes_format,
    labels_to_keep=None,
    transform_label_map=None,
    first_n_paths=None,
):
    assert len(image_paths) == len(labels_paths)
    if first_n_paths is None:
        first_n_paths = len(image_paths)
    samples = p_map(
        lambda sample_data: create_image_bboxes_detection_sample(**sample_data),
        create_bboxes_detection_samples_data_from_roheboam_format(
            image_paths[:first_n_paths],
            labels_paths[:first_n_paths],
            bboxes_format,
            transform_to_bboxes_format,
            labels_to_keep,
            transform_label_map,
        ),
    )
    return np.array(samples)


lookup = {
    "roheboam_sample_data_paths_from_path": roheboam_sample_data_paths_from_path,
    "load_bbox_from_path_roheboam_format": load_bboxes_from_path_roheboam_format,
    "load_label_from_path_roheboam_format": load_label_from_path_roheboam_format,
    "create_bboxes_detection_samples_from_roheboam_format": create_bboxes_detection_samples_from_roheboam_format,
    "create_bboxes_detection_samples_data_from_roheboam_format": create_bboxes_detection_samples_data_from_roheboam_format,
}



from textwrap import dedent

import yaml


def create_default_istio_gateway_manifest():
    manifest = yaml.safe_load(
        dedent(
            f"""
        apiVersion: networking.istio.io/v1alpha3
        kind: Gateway
        metadata:
          name: seldon-gateway
          namespace: istio-system
        spec:
          selector:
            istio: ingressgateway # use istio default controller
          servers:
          - port:
              number: 80
              name: http
              protocol: HTTP
            hosts:
            - "*"
    """
        )
    )

    return manifest



import torch


def model_sizes(m, size=(256, 256), full=True):
    "Pass a dummy input through the model to get the various sizes."
    from .hooks import hook_outputs

    hooks = hook_outputs(m)
    ch_in = in_channels(m)
    x = torch.zeros(1, ch_in, *size)
    x = m.eval()(x)
    res = [o.stored.shape for o in hooks]
    if not full:
        hooks.remove()
    return res, x, hooks if full else res


def in_channels(m):
    "Return the shape of the first weight layer in `m`."
    for layer in flatten_model(m):
        if hasattr(layer, "weight"):
            return layer.weight.shape[1]
    raise Exception("No weight layer")


def flatten_model(m):
    return sum(map(flatten_model, m.children()), []) if num_children(m) else [m]


def num_children(m):
    return len(children(m))


def children(m):
    return list(m.children())


lookup = {
    "model_sizes": model_sizes,
    "in_channels": in_channels,
    "flatten_model": flatten_model,
    "num_children": num_children,
    "children": children,
}



from copy import deepcopy
from functools import reduce
from pathlib import Path

import yaml

from ..logger import logger
from .constants import RESOURCE_KEY, VARIABLE_KEY
from .yaml_loader import RoheboamYamlDumper, RoheboamYamlLoader, Variable


def load_config_from_strings(config_strings, with_variable_replacement=False):
    configs = [
        load_config_from_string(s, with_variable_replacement=False)
        for s in config_strings
    ]
    merged_configs = merge_dicts(configs)

    if with_variable_replacement:
        merged_configs = replace_config_variables(
            merged_configs, resource_key=RESOURCE_KEY, variable_key=VARIABLE_KEY
        )

    return merged_configs


def load_config_from_string(config_string, with_variable_replacement=False):
    config = yaml.load(config_string, Loader=RoheboamYamlLoader)
    if with_variable_replacement:
        config = replace_config_variables(
            config, resource_key=RESOURCE_KEY, variable_key=VARIABLE_KEY
        )
    return config


def load_config_from_path(path, with_variable_replacement=False):
    with Path(path).open("r") as f:
        config = load_config_from_string(f, with_variable_replacement)
    return config


def load_config_from_paths(paths, with_variable_replacement=False):

    configs = [load_config_from_path(p, with_variable_replacement=False) for p in paths]
    merged_configs = merge_dicts(configs)

    if with_variable_replacement:
        merged_configs = replace_config_variables(
            merged_configs, resource_key=RESOURCE_KEY, variable_key=VARIABLE_KEY
        )
    return merged_configs


def dump_config_to_string(config):
    return yaml.dump(config, Dumper=RoheboamYamlDumper, default_flow_style=False)


def dump_config_to_path(config, save_path):
    path = Path(save_path)
    path.parent.mkdir(exist_ok=True, parents=True)
    with path.open("w") as yaml_file:
        yaml.dump(
            config, yaml_file, Dumper=RoheboamYamlDumper, default_flow_style=False
        )


def replace_config_variables(
    config, resource_key=RESOURCE_KEY, variable_key=VARIABLE_KEY
):
    if variable_key not in config:
        logger.warning("No variables passed to pipeline")
        return config

    config = deepcopy(config)
    replaced_resources = replace_variables(config[resource_key], config[variable_key])
    config[resource_key] = replaced_resources
    return config


def replace_variables(resources, variables):
    if isinstance(resources, dict):
        for name, resource in resources.items():
            resources[name] = replace_variables(resource, variables)
    elif isinstance(resources, list):
        for i, resource in enumerate(resources):
            resources[i] = replace_variables(resource, variables)
    elif isinstance(resources, Variable):
        try:
            if resources.is_nested:
                resources = replace_variable_with_group(resources, variables)
            else:
                resources = replace_variable_without_group(resources, variables)
        except Exception as e:
            logger.error(
                f"See if there are any nodes using variable: {resources.variable_group}.{resources.variable_name}"
            )
            raise e

    else:
        return resources
    return resources


def replace_variable_with_group(resource, variables):
    for i, subgroup_name in enumerate(resource.variable_groups):
        if i == 0:
            variable = variables[subgroup_name]
        else:
            variable = variable[subgroup_name]
    return variable[resource.variable_name]


def replace_variable_without_group(resources, variables):
    return variables[resources.variable_name]


def merge_dicts(dicts):
    def merge(d1, d2):
        def _merge(d1, d2, path=None):
            # Keep function pure by creating new instance
            #     d1 = dict(d1)
            if path is None:
                path = []
            for key in d2:
                if key in d1:
                    if isinstance(d1[key], dict) and isinstance(d2[key], dict):
                        _merge(d1[key], d2[key], path + [str(key)])
                    elif d1[key] == d2[key]:
                        pass  # same leaf value
                    else:
                        raise Exception("Conflict at %s" % ".".join(path + [str(key)]))
                else:
                    d1[key] = d2[key]
            return d1

        return _merge(dict(d1), d2)

    return reduce(merge, dicts)



from .core import BaseLoss, LossWrapper
from .focal_loss import CrossEntropyLoss, FocalLoss, focal_loss
from .mock_loss import MockSingleLoss
from .torchvision_fasterrcnn_loss import TorchVisionFasterRCNNLoss
from .yolo_loss import YOLOv4Loss, yolo_v4_loss

lookup = {
    "BaseLoss": BaseLoss,
    "LossWrapper": LossWrapper,
    "FocalLoss": FocalLoss,
    "CrossEntropyLoss": CrossEntropyLoss,
    "TorchVisionFasterRCNNLoss": TorchVisionFasterRCNNLoss,
    "MockSingleLoss": MockSingleLoss,
    "YOLOv4Loss": YOLOv4Loss,
    "yolo_v4_loss": yolo_v4_loss,
    "focal_loss": focal_loss,
}



import ast
import inspect
import textwrap
from collections import defaultdict
from uuid import uuid4

from .constants import RESOURCE_KEY, VARIABLE_KEY
from .yaml_loader import Reference, Variable


class ConfigurationNodeRecorder:
    def __init__(self):
        self.recorded_nodes = defaultdict(list)

    @property
    def config(self):
        config = {VARIABLE_KEY: {}, RESOURCE_KEY: {}}
        variables = {}
        for nodes in self.recorded_nodes.values():
            for node in nodes:
                # add the node
                config[RESOURCE_KEY][node.name] = node.config_properties

                # add the variables
                for variable, value in node.variables.items():
                    if variable.is_nested:
                        variable_dict_for_node = {}
                        current_nested_dict = variable_dict_for_node
                        for group_name in variable.variable_groups:
                            current_nested_dict[group_name] = {}
                            current_nested_dict = current_nested_dict[group_name]
                        current_nested_dict[variable.variable_name] = value
                        variables = {**variable_dict_for_node, **variables}
                    else:
                        variables[variable.variable_name] = value
        config[VARIABLE_KEY] = variables
        return config

    @property
    def lookup(self):
        lookup = {}
        for nodes in self.recorded_nodes.values():
            for node in nodes:
                lookup[node.callable_obj.__name__] = node.callable_obj
        return lookup

    def reset_recorded_nodes(self):
        self.recorded_nodes = defaultdict(list)

    def record(self, callable_obj, name=None, tags=[]):
        if name is None:
            name = f"{callable_obj.__name__}_{len(self.recorded_nodes[name])}"
        else:
            name = f"{name}_{self.recorded_nodes[name]}"
        configuration_node = ConfigurationNode(callable_obj, name, tags)
        self.recorded_nodes[name].append(configuration_node)
        return configuration_node


class ConfigurationNode:
    def __init__(self, callable_obj, name=None, tags=[]):
        self.callable_obj = callable_obj
        self.name = name
        self.tags = tags
        self.arguments = {}
        self.variables = {}
        self.output_names = []

    @property
    def config_properties(self):
        return {
            "properties": {
                "pointer": self.callable_obj.__name__,
                "arguments": self.arguments,
                "output_names": self.output_names,
                "tags": self.tags,
            }
        }

    def __call__(self, *args, **kwargs):
        sig = inspect.signature(self.callable_obj)
        positional_or_keyword_args = [
            param
            for param in sig.parameters.values()
            if param.kind == param.POSITIONAL_OR_KEYWORD
            or param.kind == param.POSITIONAL_ONLY
        ]
        n_positional_or_keyword_args = len(positional_or_keyword_args)

        for arg, param in zip(
            args, positional_or_keyword_args[:n_positional_or_keyword_args]
        ):
            if self._check_if_variable_is_patched(arg):
                arg_node_name = arg.node_name
                arg_output_name = arg.output_name
                self.arguments[param.name] = Reference(
                    ref_node_name=arg_node_name, output_name=arg_output_name
                )
            else:
                self.arguments[param.name] = Variable(
                    variable_group=self.name, variable_name=param.name
                )
                self.variables[
                    Variable(variable_group=self.name, variable_name=param.name)
                ] = arg

        return_values = self.callable_obj(*args, **kwargs)
        output_names = create_output_names(self.callable_obj)
        patched_outputs = []

        for output_name, return_value in zip(output_names, [return_values]):
            patched_outputs.append(
                patch_output_value(
                    return_value,
                    output_name=output_name,
                    node_name=self.name,
                )
            )

        self.output_names = output_names
        return patched_outputs[0] if len(patched_outputs) == 1 else patched_outputs

    def _check_if_variable_is_patched(self, variable):
        try:
            variable.node_name
            variable.output_name
            return True
        except Exception:
            return False


def patch_output_value(value, output_name, node_name):
    new_type = type(
        f"{str(uuid4())}",
        (type(value),),
        {
            "output_name": output_name,
            "node_name": node_name,
        },
    )
    patched_value = new_type(value)
    return patched_value


def create_output_names(f):
    class Visitor(ast.NodeVisitor):
        def visit_Return(self, node):
            self.output_names = []
            if isinstance(node.value, ast.Tuple):
                for i, _ in enumerate(node.value.elts):
                    self.output_names.append(str(i))
            else:
                self.output_names.append("0")
            self.generic_visit(node)

    visitor = Visitor()
    visitor.visit(ast.parse(textwrap.dedent(inspect.getsource(f))))
    return visitor.output_names



from enum import Enum
from pathlib import Path

import torch

from ..callback import Callback
from ..handler import CallbackHandlerContextValues


class ModelSaver(Callback):
    def __init__(
        self,
        save_directory=None,
        save_frequency="IMPROVEMENT",
        save_condition="LESSER",
        monitor_value="VALIDATION_LOSS",
        config=None,
        name=None,
    ):
        assert save_frequency in ["EVERY", "IMPROVEMENT"]
        assert save_condition in ["LESSER", "GREATER"]
        super().__init__(name)
        self.config = config
        self.save_directory = save_directory
        self.previous_monitor_value = None
        self.save_frequency = save_frequency
        self.save_condition = save_condition
        self.monitor_value = CallbackHandlerContextValues[monitor_value]

    def on_train_epoch_end(self, context):
        self.current_monitor_value = self._get_monitor_value(context)
        if (
            (context.current_epoch == 0)
            or (self.save_frequency == "EVERY")
            or (
                self.save_frequency == "IMPROVEMENT"
                and self._is_better_than_previous_monitor_value()
            )
        ):

            self._save(context)
            self.previous_monitor_value = self._get_monitor_value(context)

    def _save(self, context):
        model = context.model
        save_directory = self._construct_save_directory(context)
        model_save_path = self._save_model(save_directory, model)
        self._save_configuration(save_directory, model_save_path)

    def _construct_save_directory(self, context):
        return (
            self.save_directory
            if self.save_directory is not None
            else context.save_directory
        )

    def _save_model(self, save_directory, model):
        save_directory.mkdir(parents=True, exist_ok=True)
        model_save_path = save_directory / "model.p"
        torch.save(model.state_dict(), model_save_path)
        return model_save_path

    def _save_configuration(self, save_directory, model_save_path):
        self.config.model_save_path = model_save_path
        self.config.dump_config_to_path(Path(save_directory) / "config.yml")

    def _is_better_than_previous_monitor_value(self):
        if self.save_condition == "LESSER":
            return self.current_monitor_value < self.previous_monitor_value
        if self.save_condition == "GREATER":
            return self.current_monitor_value > self.previous_monitor_value

    def _get_monitor_value(self, context):
        return context[self.monitor_value]



import importlib

import albumentations
from albumentations import Compose
from albumentations.augmentations.crops.transforms import __all__ as albumentation_crop_transforms
from albumentations.augmentations.geometric.resize import __all__ as albumentation_geometric_resize
from albumentations.augmentations.geometric.rotate import __all__ as albumentation_geometric_rotate
from albumentations.augmentations.geometric.transforms import __all__ as albumentation_geometric_transforms
from albumentations.augmentations.transforms import __all__ as albumentation_transforms

# from albumentations.augmentations.geometric.functional import (
#     __all__ as albumentation_geometric_functional,
# )


def create_augmentation_fn(aug_fns, **kwargs):
    return Compose(aug_fns, p=1, **kwargs)


def import_albumentation_submodule(submodule_functions, submodule_string):
    return {
        name: getattr(importlib.import_module(submodule_string), name)
        for name in submodule_functions
    }


albumentation_lookup = {
    "Compose": albumentations.Compose,
    "BboxParams": albumentations.BboxParams,
    "create_augmentation_fn": create_augmentation_fn,
    **import_albumentation_submodule(
        albumentation_transforms, "albumentations.augmentations.transforms"
    ),
    **import_albumentation_submodule(
        albumentation_crop_transforms, "albumentations.augmentations.crops.transforms"
    ),
    # **import_albumentation_submodule(
    #     albumentation_geometric_functional,
    #     "albumentations.augmentations.geometric.functional",
    # ),
    **import_albumentation_submodule(
        albumentation_geometric_rotate, "albumentations.augmentations.geometric.rotate"
    ),
    **import_albumentation_submodule(
        albumentation_geometric_resize, "albumentations.augmentations.geometric.resize"
    ),
    **import_albumentation_submodule(
        albumentation_geometric_transforms,
        "albumentations.augmentations.geometric.transforms",
    ),
}



import time
from contextlib import contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory

import cloudpickle


def create_time_stamped_save_path(root_save_directory):
    current_time = f"{time.strftime('%Y%m%d-%H%M%S')}"
    time_stamped_save_path = Path(root_save_directory) / current_time
    return time_stamped_save_path


@contextmanager
def save_to_temporary_file(save_fn, name=None):
    try:
        if name is not None:
            temp_dir = TemporaryDirectory()
            save_path = Path(temp_dir.name) / name
            save_fn(save_path)
            yield str(save_path)
        else:
            temp_file = NamedTemporaryFile()
            save_fn(temp_file.name)
            yield temp_file.name
    finally:
        pass


def dump_object_to_path(path, object):
    return cloudpickle.dump(object, open(path, "wb"))


lookup = {
    "create_time_stamped_save_path": create_time_stamped_save_path,
    "save_to_temporary_file": save_to_temporary_file,
}



from functools import partial

import cv2
import numpy as np
import torch

from ....utils.training import make_one_hot
from .io import *
from .io import lookup as io_lookup
from .plot import *
from .plot import lookup as plot_lookup
from .tile_with_square import *
from .tile_with_square import lookup as tile_with_square_lookup


def image_to_tensor(
    image, normalize_image_fn=None, scale_to_range_fn=None, normalize_tensor_fn=None
):
    """Transforms an image to a tensor

    :param image: image
    :type image: numpy.array
    :param normalize_image_fn: function to normalize image, look at: pytorch_toolbox.vision.utils.normalization.normalize_image, defaults to None
    :type normalize_image_fn: Callable, optional
    :param scale_to_range_fn: function to transform an image to a certain range, look at: pytorch_toolbox.vision.utils._utils.scale_to_range,
    defaults to None
    :type scale_to_range_fn: Callable, optional
    :param normalize_tensor_fn: function to normalize the tensor, look at: pytorch_toolbox.vision.utils.normalization.normalize, defaults to None
    :type normalize_tensor_fn: Callable, optional
    :return: Tensor
    :rtype: torch.Tensor
    """

    n_dims = len(image.shape)
    assert len(image.shape) in [
        2,
        3,
    ], "There must be two, or three dimensions to the image"
    if n_dims == 2:
        image = np.expand_dims(image, -1)

    if image.dtype != np.float32:
        image = image.astype(np.float32)

    if normalize_image_fn is not None:
        image = normalize_image_fn(image)

    if scale_to_range_fn is not None:
        image = scale_to_range_fn(image)

    tensor = numpy_image_to_tensor(image)

    if normalize_tensor_fn is not None:
        tensor = normalize_tensor_fn(tensor)

    return tensor


def tensor_to_image(
    tensor,
    unnormalize_image_fn=None,
    unscale_to_range_fn=None,
    unnormalize_tensor_fn=None,
    convert_to_image_type_fn=None,
):
    assert len(tensor.shape) in [
        3,
        4,
    ], "There must be three dimensions in the tensor C x H x W or B x C x H x W"
    if unnormalize_tensor_fn is not None:
        tensor = unnormalize_tensor_fn(tensor)

    image = tensor_to_numpy_image(tensor)

    if image.dtype != np.float32:
        image = image.astype(np.float32)

    if unscale_to_range_fn is not None:
        image = unscale_to_range_fn(image)

    if unnormalize_image_fn is not None:
        image = unnormalize_image_fn(image)

    if convert_to_image_type_fn is not None:
        image = convert_to_image_type_fn(image)

    return image


def unit_normalized_tensor_to_8_bit_image(tensor):
    return tensor_to_image(
        tensor,
        unscale_to_range_fn=partial(scale_to_range, original_max_val=1, max_val=255),
        convert_to_image_type_fn=to_uint8_image,
    )


def numpy_image_to_tensor(image):
    if len(image.shape) == 3:
        image = np.transpose(image, (2, 0, 1))
    elif len(image.shape) == 4:
        image = np.transpose(image, (2, 3, 1))
    tensor = torch.from_numpy(image)
    return tensor


def tensor_to_numpy_image(tensor):
    image = tensor.cpu().detach().numpy()
    if len(tensor.shape) == 3:
        image = np.transpose(image, (1, 2, 0))
    elif len(tensor.shape) == 4:
        image = np.transpose(image, (0, 2, 3, 1))

    return image


def is_rgb(image):
    return len(image.shape) == 3 and image.shape[-1] == 3


def to_rgb(image):
    if is_rgb(image):
        return image
    else:
        # Catch case where we have (H x W x 1)
        image = image.squeeze()
    return np.stack([image] * 3, axis=-1)


def to_uint8_image(image, clip=True):
    image = np.array(image)
    if clip:
        image = np.clip(image, a_min=0, a_max=255)
    return image.astype(np.uint8)


def scale_to_range(
    image, original_min_val=0, min_val=0, original_max_val=255, max_val=1
):
    scaled_image = (max_val - min_val) * (image - min_val) / (
        original_max_val - original_min_val
    ) + min_val
    return scaled_image


def create_mask_from_vertices(vertices, image_width=None, image_height=None):
    """Create mask from a list of vertices

    :param vertices: list of vertices in OpenCV format
    :type vertices: [[x1, y2], [x2, y2], ...]
    :param image_width: [description], defaults to None
    :type image_width: [type], optional
    :param image_height: [description], defaults to None
    :type image_height: [type], optional
    :return: [description]
    :rtype: [type]
    """
    xs = [p[0] for p in vertices]
    ys = [p[1] for p in vertices]

    if image_width is None:
        image_width = max(xs) + 1

    if image_height is None:
        image_height = max(ys) + 1

    vertices = np.array([vertices], dtype=np.int32)
    img = np.zeros([image_height, image_width], dtype=np.uint8)
    cv2.fillPoly(img, vertices, 1)

    return img


def calculate_pixels_in_mask(mask, value=None):
    if value is None:
        return len(mask[mask > 0])
    else:
        return len(mask[mask == value])


def load_mask_from_grayscale_and_transform_to_one_hot(n_classes):
    return lambda mask_path: make_one_hot(
        imread_grayscale(mask_path), n_classes=n_classes
    )


def mask_to_tensor(mask, make_one_hot_fn=None):
    if make_one_hot_fn is not None:
        mask = mask.squeeze()
        mask = make_one_hot_fn(mask)

    tensor = numpy_image_to_tensor(mask)
    return tensor


def tensor_to_mask(tensor, reduce_one_hot_fn=None, convert_to_image_type_fn=None):
    image = tensor_to_numpy_image(tensor)
    if reduce_one_hot_fn is not None:
        image = reduce_one_hot_fn(image)
    if convert_to_image_type_fn is not None:
        image = convert_to_image_type_fn(image)
    return image


def reduce_max_with_dim(array, dim):
    return np.argmax(array, axis=dim)


def set_only_class(mask, only_class=None, binary_mask=True):
    if only_class is None:
        return mask
    mask[mask != only_class] = 0
    if binary_mask:
        mask[mask == only_class] = 1
    return mask


lookup = {
    "calculate_pixels_in_mask": calculate_pixels_in_mask,
    "create_mask_from_vertices": create_mask_from_vertices,
    "image_to_tensor": image_to_tensor,
    "is_rgb": is_rgb,
    "load_mask_from_grayscale_and_transform_to_one_hot": load_mask_from_grayscale_and_transform_to_one_hot,
    "mask_to_tensor": mask_to_tensor,
    "tensor_to_mask": tensor_to_mask,
    "reduce_max_with_dim": reduce_max_with_dim,
    "numpy_image_to_tensor": numpy_image_to_tensor,
    "tensor_to_image": tensor_to_image,
    "tensor_to_numpy_image": tensor_to_numpy_image,
    "to_rgb": to_rgb,
    "to_uint8_image": to_uint8_image,
    "scale_to_range": scale_to_range,
    "unit_normalized_tensor_to_8_bit_image": unit_normalized_tensor_to_8_bit_image,
    "set_only_class": set_only_class,
    **io_lookup,
    **plot_lookup,
    **tile_with_square_lookup,
}



from pathlib import Path

from pytorch_lightning.callbacks import ModelCheckpoint

from ....core.callbacks import Callback


class PyTorchLightningModelCheckpoint(Callback):
    def __init__(self, config, root_save_path):
        super().__init__()
        self.config = config
        self.model_save_path = Path(root_save_path) / "checkpoints"
        self.model_checkpoint = ModelCheckpoint(dirpath=str(self.model_save_path))

    @property
    def current_model_path(self):
        return self.model_save_path / self.current_model_save_name

    @property
    def best_model_path(self):
        return self.model_checkpoint.best_model_path

    @property
    def best_config_path(self):
        return f"{self.model_checkpoint.best_model_path}.config.yml"

    def on_validation_end(self, context):
        trainer = context.trainer
        epoch, step = trainer.current_epoch, trainer.global_step
        self.current_model_save_name = Path(
            self.format_checkpoint_name(epoch, step)
        ).name
        self.model_save_path.mkdir(exist_ok=True, parents=True)
        self.model_checkpoint.save_checkpoint(trainer)
        self._save_configuration()

    def format_checkpoint_name(self, epoch, step):
        return str(self.model_save_path / f"epoch={epoch}-step={step}.ckpt")

    def _save_configuration(self):
        config_save_path = (
            self.model_save_path / f"{self.current_model_save_name}.config.yml"
        )
        self.config.model_save_path = str(
            self.model_save_path / self.current_model_save_name
        )
        self.config.save_path = config_save_path
        self.config.dump_config_to_path(config_save_path)

    def on_save_checkpoint(self, trainer_callback_hook_mixin, model):
        return None


lookup = {"PyTorchLightningModelCheckpoint": PyTorchLightningModelCheckpoint}



from pathlib import Path

import numpy as np
from p_tqdm import p_map

from .....vision.utils.image import imread_rgb
from ..data import create_image_classification_sample


def create_image_classification_samples_data_for_label_per_folder(
    root_label_folder, label_map, load_image_fn=imread_rgb
):
    arguments = []
    for label_folder in Path(root_label_folder).glob("*"):
        for image_path in Path(label_folder).glob("*"):
            arguments.append(
                {
                    "image_path": image_path,
                    "load_image_fn": load_image_fn,
                    "label_data": label_map[label_folder.name],
                }
            )

    return np.array(arguments)


def create_labels_from_sample_arguments_for_label_per_folder(sample_arguments):
    return sample_arguments, [s["label_data"] for s in sample_arguments]


def create_labels_from_samples(samples):
    return [s.label.data for s in samples]


def create_image_classification_samples_for_label_per_folder(
    root_label_folder, label_map, load_image_fn=imread_rgb
):
    samples = p_map(
        _create_image_classification_sample_lambda,
        create_image_classification_samples_data_for_label_per_folder(
            root_label_folder, label_map, load_image_fn
        ),
    )
    return np.array(samples)


def _create_image_classification_sample_lambda(sample_data):
    return create_image_classification_sample(**sample_data)


lookup = {
    "create_labels_from_samples": create_labels_from_samples,
    "create_image_classification_samples_for_label_per_folder": create_image_classification_samples_for_label_per_folder,
    "create_image_classification_samples_data_for_label_per_folder": create_image_classification_samples_data_for_label_per_folder,
    "create_labels_from_sample_arguments_for_label_per_folder": create_labels_from_sample_arguments_for_label_per_folder,
}



from abc import ABC
from uuid import uuid4


class Callback(ABC):
    _order = 0

    def __init__(self, name=None):
        if name is None:
            self.name = f"{self.__class__.__name__}-{str(uuid4())[:4]}"
        else:
            self.name = name

    def on_setup(self, context):
        pass

    def on_init_begin(self, context):
        pass

    def on_init_end(self, context):
        pass

    def on_train_begin(self, context):
        pass

    def on_train_epoch_begin(self, context):
        pass

    def on_train_batch_begin(self, context):
        pass

    def on_train_loss_begin(self, context):
        pass

    def on_backward_begin(self, context):
        pass

    def on_backward_end(self, context):
        pass

    def on_train_step_end(self, context):
        pass

    def on_train_batch_end(self, context):
        pass

    def on_validation_begin(self, context):
        pass

    def on_validation_epoch_begin(self, context):
        pass

    def on_validation_batch_begin(self, context):
        pass

    def on_validation_loss_begin(self, context):
        pass

    def on_validation_step_end(self, context):
        pass

    def on_validation_batch_end(self, context):
        pass

    def on_validation_epoch_end(self, context):
        pass

    def on_validation_end(self, context):
        pass

    def on_train_epoch_end(self, context):
        pass

    def on_train_end(self, context):
        pass

    def on_test_begin(self, context):
        pass

    def on_test_epoch_begin(self, context):
        pass

    def on_test_batch_begin(self, context):
        pass

    def on_test_loss_begin(self, context):
        pass

    def on_test_step_end(self, context):
        pass

    def on_test_batch_end(self, context):
        pass

    def on_test_epoch_end(self, context):
        pass

    def on_test_end(self, context):
        pass

    def on_teardown(self, context):
        pass



import functools
from enum import Enum
from pprint import pformat
from textwrap import dedent

from ..logger import logger
from .constants import SHOULD_REMOVE_ARGUMENT
from .graph_construction import Reference


class NodeRemoveMode(Enum):
    AUTO = 0
    ONLY = 1
    PROPAGATE = 2


class Node:
    def __init__(
        self,
        name,
        references,
        pointer,
        partial,
        arguments,
        output_names,
        tags=[],
        should_remove=False,
        remove_mode=NodeRemoveMode.PROPAGATE,
    ):
        self.name = name
        self.references = references
        self.referenced_by = []
        self.pointer = pointer
        self.partial = partial
        self.arguments = arguments
        self.output_names = output_names
        self.tags = tags
        self.should_remove = should_remove
        self.remove_mode = remove_mode
        self.reference_replaced_arguments = {}
        self.output = None

    @staticmethod
    def load_properties_with_default_values(properties, lookups):
        pointer_name = properties["pointer"]
        assert (
            pointer_name in lookups
        ), f"There is no pointer called {pointer_name} in {pformat(lookups)}"
        return {
            "pointer": lookups[pointer_name],
            "partial": properties.get("partial", False),
            "arguments": properties.get("arguments", {}),
            "output_names": properties.get("output_names", []),
            "should_remove": properties.get("should_remove", False),
            "remove_mode": properties.get("remove_mode", NodeRemoveMode.AUTO),
            "tags": properties.get("tags", []),
        }

    def __repr__(self):
        return dedent(
            f"""
            Name: {self.name}
            Pointer: {self.pointer}
            Arguments: {pformat(self.arguments)}
            Reference Replaced Arguments: {pformat(self.reference_replaced_arguments)}
            Should Remove: {self.should_remove}
            Remove Mode: {self.remove_mode}
            Tags: {self.tags},
        """
        )

    def create_output(self):
        assert not isinstance(
            self.output_names, str
        ), f"The output_names for node: {self.name} can't be a string, only a list"
        self._call_pointer_and_set_output()

    def _call_pointer_and_set_output(self):
        if self.partial:
            assert (
                len(self.output_names) == 1
            ), f"If the output of node: {self.name} is partial, then there should be one output, {len(self.output_names)} outputs are found"
            self.output = {
                self.output_names[0]: functools.partial(
                    self.pointer, **self.reference_replaced_arguments
                )
            }
        else:
            try:
                output = self.pointer(**self.reference_replaced_arguments)
            except TypeError as e:
                logger.error(e)
                logger.error(f"This error was produced when calling node:\n{self}")
                logger.error(e, exc_info=True)
                raise e

            if output is not None:
                assert (
                    len(self.output_names) > 0
                ), f"There are no output_names defined in properties for the node: {self.name}"
                iterable_output = [output] if len(self.output_names) == 1 else output
                self.output = {
                    output_name: output_value
                    for output_name, output_value in zip(
                        self.output_names, iterable_output
                    )
                }


def remove_node_arguments(arguments, nodes_to_remove):
    if isinstance(arguments, dict):
        name_of_arguments_to_remove = []
        for name, value in arguments.items():
            updated_argument = remove_node_arguments(value, nodes_to_remove)
            if updated_argument == SHOULD_REMOVE_ARGUMENT:
                name_of_arguments_to_remove.append(name)
        for name in name_of_arguments_to_remove:
            del arguments[name]
    elif isinstance(arguments, list):
        values_to_remove = []
        for value in arguments:
            updated_argument = remove_node_arguments(value, nodes_to_remove)
            if updated_argument == SHOULD_REMOVE_ARGUMENT:
                values_to_remove.append(value)
        for value in values_to_remove:
            arguments.remove(value)
    elif isinstance(arguments, Reference):
        should_remove = arguments.ref_node_name in [
            node_to_remove.name for node_to_remove in nodes_to_remove
        ]
        if should_remove:
            return SHOULD_REMOVE_ARGUMENT
    return arguments



import docker

from ....engine.utils.convenience import run_shell_command


def stream_docker_logs(log_generator):
    while True:
        try:
            output = log_generator.__next__()
            if "stream" in output:
                output_str = output["stream"].strip("\r\n").strip("\n")
                print(output_str)
        except StopIteration:
            print("Docker image build complete.")
            break
        except ValueError:
            print(f"Error parsing output from docker image build: {output}")


def remove_image(image, force=True, no_prune=False):
    client = docker.from_env()
    client.images.remove(image=image, force=force, noprune=no_prune)


def build_image(image_tag, context_path, docker_file_path=None):
    # client = docker.from_env()
    # image, log_generator = client.images.build(path=docker_file_path, tag=image_tag)
    # stream_docker_logs(log_generator)

    # Have to use this workaround or the layers won't cache for some reason...
    build_command = ["docker", "build", str(context_path), "-t", str(image_tag)]

    if docker_file_path is not None:
        build_command += ["-f", str(docker_file_path)]
    run_shell_command(" ".join(build_command))



from pathlib import Path
from uuid import uuid4

import matplotlib.pyplot as plt
import torch.utils

from ......core.data import Data
from ......vision.utils.image import imread_grayscale, imread_rgb, plot_mask_on_image, to_uint8_image


class ImageSegmentationSample:
    """

    If label is not given, then it will be generated by taking the max value of each mask

    :return: [description]
    :rtype: [type]
    """

    def __init__(self, image, mask=None, weight_map=None, name=None):
        self.image = image
        self.mask = mask
        self.weight_map = weight_map
        self.name = str(uuid4()) if name is None else name

    @classmethod
    def create(
        cls,
        image_data=None,
        image_path=None,
        load_image_fn=imread_rgb,
        mask_data=None,
        mask_path=None,
        load_mask_fn=imread_grayscale,
        weight_map_data=None,
        weight_map_path=None,
        load_weight_map_fn=None,
        name=None,
    ):
        if name is None:
            name = Path(image_path).stem if image_path is not None else str(uuid4())

        image = Data(image_data, image_path, load_image_fn, name=name)

        if mask_data is None and mask_path is None:
            mask = None
        else:
            mask = Data(mask_data, mask_path, load_mask_fn, name=name)

        if weight_map_data is None and weight_map_path is None:
            weight_map = None
        else:
            weight_map = Data(
                weight_map_data, weight_map_path, load_weight_map_fn, name=name
            )

        return cls(image, mask, weight_map, name)

    @property
    def is_inference(self):
        return self.has_masks is False and self.has_weight_map is False

    @property
    def has_mask(self):
        return self.mask is not None

    @property
    def has_weight_map(self):
        return self.weight_map is not None

    @property
    def data(self):
        return (
            self.image.data,
            self.mask.data if self.has_mask else None,
            self.weight_map.data if self.has_weight_map else None,
            self.name,
        )

    def plot(self, overlay_mask_with_image=True, figsize=(18, 18)):
        n_rows = 1
        n_cols = (
            1
            + int(self.has_mask)
            + int(self.has_weight_map)
            + int(overlay_mask_with_image)
        )
        _, axs = plt.subplots(n_rows, n_cols, figsize=figsize)

        plot_idx = 0
        axs[plot_idx].imshow(self.image.data)
        axs[plot_idx].set_title(self.name)
        plot_idx += 1

        if self.has_mask:
            axs[plot_idx].imshow(to_uint8_image(self.mask.data))
            plot_idx += 1

        if overlay_mask_with_image:
            axs[plot_idx].imshow(
                plot_mask_on_image(
                    to_uint8_image(self.image.data), to_uint8_image(self.mask.data)
                )
            )
            plot_idx += 1

        if self.has_weight_map:
            axs[plot_idx].imshow(self.weight_map.data)
            plot_idx + 1


lookup = {"ImageSegmentationSample": ImageSegmentationSample}



import torch
import torch.utils

from ......vision.utils.image import imread_grayscale, imread_rgb
from .sample import ImageSegmentationSample


def create_image_segmentation_sample(
    image_data=None,
    image_path=None,
    load_image_fn=imread_rgb,
    mask_data=None,
    mask_path=None,
    load_mask_fn=imread_grayscale,
    weight_map_data=None,
    weight_map_path=None,
    load_weight_map_fn=None,
    name=None,
):
    return ImageSegmentationSample.create(
        image_data=image_data,
        image_path=image_path,
        load_image_fn=load_image_fn,
        mask_data=mask_data,
        mask_path=mask_path,
        load_mask_fn=load_mask_fn,
        weight_map_data=weight_map_data,
        weight_map_path=weight_map_path,
        load_weight_map_fn=load_weight_map_fn,
        name=name,
    )


lookup = {"create_image_segmentation_sample": create_image_segmentation_sample}



import json
from abc import ABC
from pathlib import Path

import hub
from hub.store.store import get_fs_and_path

from ...utils.convenience import git_commit_hash


class ActiveLoopHub(ABC):
    def __init__(self, activeloop_dataset):
        self.activeloop_dataset = activeloop_dataset
        self.additional_meta_info = self._load_additional_meta_info()

    @property
    def commit_id(self):
        return self.activeloop_dataset._commit_id

    @property
    def previous_commit_id(self):
        try:
            return self.all_previous_commit_ids[-1]
        except IndexError:
            print(
                "Error when returning previous commit as there this is the first commit"
            )
            return {}

    @property
    def current_node(self):
        return (
            self.activeloop_dataset._version_node.parent
            if not self.activeloop_dataset._version_node.children
            else self.activeloop_dataset._version_node
        )

    @property
    def all_previous_commit_ids(self):
        previous_commit_ids = []
        _current_node = self.current_node
        while _current_node:
            previous_commit_ids.append(_current_node.commit_id)
            _current_node = _current_node.parent
        return previous_commit_ids[::-1]

    @property
    def previous_commit_additional_meta_info(self):
        return self.all_additional_meta_info.get(self.previous_commit_id)

    @property
    def all_additional_meta_info(self):
        _, path = get_fs_and_path(self.activeloop_dataset.url)
        try:
            with open(str(Path(path) / "additional_meta.json"), "r") as f:
                return json.load(f)
        except FileNotFoundError:
            print("error loading additional info")
            return dict()

    def _store_additional_meta_info_for_commit_id(self, commit_id):
        _, path = get_fs_and_path(self.activeloop_dataset.url)
        all_additional_meta_info = self.all_additional_meta_info
        all_additional_meta_info[commit_id] = self.additional_meta_info
        with open(str(Path(path) / "additional_meta.json"), "w") as f:
            json.dump(all_additional_meta_info, f)

    def _load_additional_meta_info(self):
        self.additional_meta_info = self.all_additional_meta_info.get(
            self.commit_id, {}
        )
        return self.additional_meta_info

    def __len__(self):
        return len(self.activeloop_dataset)

    def __getitem__(self, index):
        return self.activeloop_dataset.__getitem__(index)

    def commit(self, message):
        previous_commit_id = self.activeloop_dataset.commit(message)
        self._store_additional_meta_info_for_commit_id(previous_commit_id)
        return previous_commit_id

    def save(self):
        self._store_additional_meta_info_for_commit_id(self.commit_id)
        self.activeloop_dataset.save()

    def store(self, url):
        self.activeloop_dataset = self.activeloop_dataset.store(url=url)
        return self

    def filter(self, filter_fn):
        return self.activeloop_dataset.filter(filter_fn)

    def get_index_of_id(self, id):
        return self.filter(lambda sample: sample.compute()["id"] == id).indexes[0]

    def checkout(self, address, create=False):
        self.activeloop_dataset.checkout(address, create=create)
        self._load_additional_meta_info()

    def checkout_for_training_run(self):
        self.checkout(git_commit_hash(), create=True)

    def resize_shape(self, shape):
        self.activeloop_dataset.resize_shape(shape)

    def log(self):
        return self.activeloop_dataset.log()

    def debug_logs(self):
        print("All additional meta info")
        print(self.all_additional_meta_info)

        print()
        print("Additional meta info")
        print(self.additional_meta_info)

        print()
        print("Commit id to be used")
        print(self.commit_id)

        print()
        print("Logs")
        self.log()

        print()
        print("Previous commits")
        print(self.all_previous_commit_ids)

    @property
    def branches(self):
        return self.activeloop_dataset.branches



from abc import abstractmethod

import numpy as np

from ...utils.convenience import camel2snake
from ...utils.interfaces import TransformToCallKeywordArguments


class MetricWrapper:
    def __init__(self, metrics):
        super().__init__()
        for metric in metrics:
            setattr(self, camel2snake(metric.__class__.__name__), metric)
        self.metrics = metrics

    def __call__(self, predictions, targets={}, auxillary={}):
        all_metrics = []
        for metric in self.metrics:
            metric_args = metric.transform_to_kwargs(
                predictions=predictions, targets=targets, auxillary=auxillary
            )
            all_metrics.append(metric(**metric_args))
        return all_metrics


class BaseMetric(TransformToCallKeywordArguments):
    metric_name_incrementor = 0

    def __init__(self, name):
        if name is None:
            self.name = (
                f"{self.__class__.__name__}-{BaseMetric.metric_name_incrementor}"
            )
            BaseMetric.metric_name_incrementor += 1
        else:
            self.name = name


lookup = {"MetricWrapper": MetricWrapper}



import hub
import numpy as np
from hub.schema import Image, Segmentation, Text

from .....integrations.activeloop import ActiveLoopHub, compute_activeloop_element
from ..data import create_image_segmentation_sample


class ImageSegmentationActiveLoopHub(ActiveLoopHub):
    def __init__(self, activeloop_dataset):
        super().__init__(activeloop_dataset)

    @staticmethod
    def create_schema_for_rgb(
        id_max_shape, image_max_shape, mask_max_shape, label_names
    ):
        return {
            "id": Text(shape=(None,), max_shape=id_max_shape),
            "image": Image(
                shape=(None, None, 3), dtype="uint8", max_shape=image_max_shape
            ),
            "mask": Segmentation(
                shape=(None, None),
                dtype="uint8",
                names=label_names,
                max_shape=mask_max_shape,
            ),
        }

    @classmethod
    def create_from_url_and_checkout(cls, url, address):
        activeloop_dataset = hub.Dataset(url)
        activeloop_dataset.checkout(address)
        ds = cls(activeloop_dataset)
        return ds

    def create_samples(self):
        return [
            create_image_segmentation_sample(**sample_data)
            for sample_data in self.create_samples_data()
        ]

    def create_samples_data(self):
        arguments = [
            {
                "name": data["id"].compute(),
                "image_path": data["image"],
                "load_image_fn": compute_activeloop_element,
                "mask_path": data["mask"],
                "load_mask_fn": compute_activeloop_element,
            }
            for data in self.activeloop_dataset
        ]
        return np.array(arguments)

    def create_sample_data_from_index(self, index):
        data = self.activeloop_dataset[index]
        return {
            "name": data["id"].compute(),
            "image_path": data["image"],
            "load_image_fn": compute_activeloop_element,
            "mask_path": data["mask"],
            "load_mask_fn": compute_activeloop_element,
        }

    def create_sample_from_index(self, index):
        return create_image_segmentation_sample(
            **self.create_sample_data_from_index(index)
        )

    def update_index_with_sample(self, index, sample):
        self.activeloop_dataset["id"][index] = sample.name
        self.activeloop_dataset["image"][index] = sample.image.data
        self.activeloop_dataset["mask"][index] = sample.mask.data


def create_image_segmentation_activeloop_data_from_sample(sample):
    return {"id": sample.name, "image": sample.image.data, "mask": sample.mask.data}


def create_image_segmentation_activeloop_and_checkout(url, address):
    return ImageSegmentationActiveLoopHub.create_from_url_and_checkout(url, address)


lookup = {
    "ImageSegmentationActiveLoopHub": ImageSegmentationActiveLoopHub,
    "create_image_segmentation_activeloop_data_from_sample": create_image_segmentation_activeloop_data_from_sample,
    "create_image_segmentation_activeloop_and_checkout": create_image_segmentation_activeloop_and_checkout,
}



import math

import torch
import torch.nn as nn

from . import BaseLoss


class YOLOv4Loss(BaseLoss):
    def __init__(self, strides, iou_threshold_loss=0.5, name=None):
        super().__init__(name)
        self.strides = strides
        self.iou_threshold_loss = iou_threshold_loss

    def to_call_kwargs(self, d):
        return {**d["predictions"], **d["targets"]}

    @property
    def unreduced_loss(self):
        return self._unreduced_loss

    @property
    def per_sample_loss(self):
        return self._per_sample_loss

    @property
    def reduced_loss(self):
        return self._reduced_loss

    @property
    def total_loss(self):
        return self._total_loss

    def __call__(
        self,
        predictions,
        predictions_decoded,
        label_and_bboxes_small,
        label_and_bboxes_medium,
        label_and_bboxes_large,
        small_bboxes,
        medium_bboxes,
        large_bboxes,
    ):
        loss_small_ciou, loss_small_conf, loss_small_cls = yolo_v4_loss(
            predictions[0],
            predictions_decoded[0],
            label_and_bboxes_small,
            small_bboxes,
            self.strides[0],
            self.iou_threshold_loss,
            reduce=False,
        )

        loss_medium_ciou, loss_medium_conf, loss_medium_cls = yolo_v4_loss(
            predictions[1],
            predictions_decoded[1],
            label_and_bboxes_medium,
            medium_bboxes,
            self.strides[1],
            self.iou_threshold_loss,
            reduce=False,
        )

        loss_large_ciou, loss_large_conf, loss_large_cls = yolo_v4_loss(
            predictions[2],
            predictions_decoded[2],
            label_and_bboxes_large,
            large_bboxes,
            self.strides[2],
            self.iou_threshold_loss,
            reduce=False,
        )

        self._unreduced_loss = {
            f"{self.name}_loss_small_ciou": loss_small_ciou,
            f"{self.name}_loss_small_conf": loss_small_conf,
            f"{self.name}_loss_small_cls": loss_small_cls,
            f"{self.name}_loss_medium_ciou": loss_medium_ciou,
            f"{self.name}_loss_medium_conf": loss_medium_conf,
            f"{self.name}_loss_medium_cls": loss_medium_cls,
            f"{self.name}_loss_large_ciou": loss_large_ciou,
            f"{self.name}_loss_large_conf": loss_large_conf,
            f"{self.name}_loss_large_cls": loss_large_cls,
        }

        self._per_sample_loss = {
            f"{self.name}_loss_small_ciou": self.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
                loss_small_ciou
            ),
            f"{self.name}_loss_small_conf": self.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
                loss_small_conf
            ),
            f"{self.name}_loss_small_cls": self.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
                loss_small_cls
            ),
            f"{self.name}_loss_medium_ciou": self.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
                loss_medium_ciou
            ),
            f"{self.name}_loss_medium_conf": self.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
                loss_medium_conf
            ),
            f"{self.name}_loss_medium_cls": self.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
                loss_medium_cls
            ),
            f"{self.name}_loss_large_ciou": self.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
                loss_large_ciou
            ),
            f"{self.name}_loss_large_conf": self.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
                loss_large_conf
            ),
            f"{self.name}_loss_large_cls": self.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
                loss_large_cls
            ),
        }

        self._reduced_loss = {
            loss_name: loss_values.mean()
            for loss_name, loss_values in self.per_sample_loss.items()
        }

        self._total_loss = torch.stack(list(self._reduced_loss.values())).sum()

        return self._total_loss


def yolo_v4_loss(p, p_d, label, bboxes, stride, iou_threshold_loss, reduce=True):
    """
    label is BS x G x G x N_ANCHOR x ([x,y,w,h,c] + N_CLASSES)
    bboxes is BS x 150 x 4
    (1)The loss of regression of boxes.
        GIOU loss is defined in  https://arxiv.org/abs/1902.09630.

    Note: The loss factor is 2-w*h/(img_size**2), which is used to influence the
            balance of the loss value at different scales.
    (2)The loss of confidence.
        Includes confidence loss values for foreground and background.

    Note: The backgroud loss is calculated when the maximum iou of the box predicted
            by the feature point and all GTs is less than the threshold.
    (3)The loss of classes。
        The category loss is BCE, which is the binary value of each class.

    :param stride: The scale of the feature map relative to the original image

    :return: The average loss(loss_giou, loss_conf, loss_cls) of all batches of this detection layer.
    """
    BCE = nn.BCEWithLogitsLoss(reduction="none")
    FOCAL = FocalLoss(gamma=2, alpha=1.0, reduction="none")

    batch_size, grid = p.shape[:2]
    img_size = stride * grid

    p_conf = p[..., 4:5]
    p_cls = p[..., 5:]

    p_d_xywh = p_d[..., :4]

    label_xywh = label[..., :4]
    label_obj_mask = label[..., 4:5]
    label_cls = label[..., 5:]
    #         label_mix = label[..., 5:6]

    # loss ciou, p_d_xywh are the anchors, label_xywh are the labels for the anchors,
    # usually they are [0, 0, 0, 0] if the anchor doesn't have an object associated with it
    ciou = CIOU_xywh_torch(p_d_xywh, label_xywh).unsqueeze(-1)

    # The scaled weight of bbox is used to balance the impact of small objects and large objects on loss.
    bbox_loss_scale = 2.0 - 1.0 * label_xywh[..., 2:3] * label_xywh[..., 3:4] / (
        img_size ** 2
    )
    loss_ciou = label_obj_mask * bbox_loss_scale * (1.0 - ciou)

    # loss confidence
    iou = CIOU_xywh_torch(
        p_d_xywh.unsqueeze(4), bboxes.unsqueeze(1).unsqueeze(1).unsqueeze(1)
    )
    iou_max = iou.max(-1, keepdim=True)[0]
    label_noobj_mask = (1.0 - label_obj_mask) * (iou_max < iou_threshold_loss).float()

    loss_conf = label_obj_mask * FOCAL(
        input=p_conf, target=label_obj_mask
    ) + label_noobj_mask * FOCAL(input=p_conf, target=label_obj_mask)

    # loss classes
    loss_cls = label_obj_mask * BCE(input=p_cls, target=label_cls)

    loss_ciou_reduced = (torch.sum(loss_ciou)) / batch_size
    loss_conf_reduced = (torch.sum(loss_conf)) / batch_size
    loss_cls_reduced = (torch.sum(loss_cls)) / batch_size
    loss_reduced = loss_ciou + loss_conf + loss_cls

    if reduce:
        return loss_reduced, loss_ciou_reduced, loss_conf_reduced, loss_cls_reduced
    else:
        # BS x grid_size x grid_size x n_anchors_at_scale x 1 for loss_ciou & loss_conf
        # BS x grid_size x grid_size x n_anchors_at_scale x n_classes for loss_cls
        return loss_ciou, loss_conf, loss_cls


class _YOLOv4Loss(nn.Module):
    def __init__(self, strides, iou_threshold_loss=0.5):
        super().__init__()
        self.__iou_threshold_loss = iou_threshold_loss
        self.__strides = strides

    def forward(
        self,
        predictions,
        predictions_decoded,
        label_and_bboxes_small,
        label_and_bboxes_medium,
        label_and_bboxes_large,
        small_bboxes,
        medium_bboxes,
        large_bboxes,
    ):
        """
        :param p: Predicted offset values for three detection layers.
                    The shape is [p0, p1, p2], ex. p0=[bs, grid, grid, anchors, tx+ty+tw+th+conf+cls_20]
        :param p_d: Decodeed predicted value. The size of value is for image size.
                    ex. p_d0=[bs, grid, grid, anchors, x+y+w+h+conf+cls_20]
        :param label_sbbox: Small detection layer's label. The size of value is for original image size.
                    shape is [bs, grid, grid, anchors, x+y+w+h+conf+mix+cls_20]
        :param label_mbbox: Same as label_sbbox.
        :param label_lbbox: Same as label_sbbox.
        :param sbboxes: Small detection layer bboxes.The size of value is for original image size.
                        shape is [bs, 150, x+y+w+h]
        :param mbboxes: Same as sbboxes.
        :param lbboxes: Same as sbboxes
        """
        strides = self.__strides
        loss_small_ciou, loss_small_conf, loss_small_cls = self.__cal_loss_per_layer(
            predictions[0],
            predictions_decoded[0],
            label_and_bboxes_small,
            small_bboxes,
            strides[0],
            reduce=False,
        )
        loss_medium_ciou, loss_medium_conf, loss_medium_cls = self.__cal_loss_per_layer(
            predictions[1],
            predictions_decoded[1],
            label_and_bboxes_medium,
            medium_bboxes,
            strides[1],
            reduce=False,
        )
        loss_large_ciou, loss_large_conf, loss_large_cls = self.__cal_loss_per_layer(
            predictions[2],
            predictions_decoded[2],
            label_and_bboxes_large,
            large_bboxes,
            strides[2],
            reduce=False,
        )

        return (
            [loss_small_ciou, loss_small_conf, loss_small_cls],
            [loss_medium_ciou, loss_medium_conf, loss_medium_cls],
            [loss_large_ciou, loss_large_conf, loss_large_cls],
        )

    def __cal_loss_per_layer(self, p, p_d, label, bboxes, stride, reduce=True):
        """
        label is BS x G x G x N_ANCHOR x ([x,y,w,h,c] + N_CLASSES)
        bboxes is BS x 150 x 4
        (1)The loss of regression of boxes.
          GIOU loss is defined in  https://arxiv.org/abs/1902.09630.

        Note: The loss factor is 2-w*h/(img_size**2), which is used to influence the
             balance of the loss value at different scales.
        (2)The loss of confidence.
            Includes confidence loss values for foreground and background.

        Note: The backgroud loss is calculated when the maximum iou of the box predicted
              by the feature point and all GTs is less than the threshold.
        (3)The loss of classes。
            The category loss is BCE, which is the binary value of each class.

        :param stride: The scale of the feature map relative to the original image

        :return: The average loss(loss_giou, loss_conf, loss_cls) of all batches of this detection layer.
        """
        BCE = nn.BCEWithLogitsLoss(reduction="none")
        FOCAL = FocalLoss(gamma=2, alpha=1.0, reduction="none")

        batch_size, grid = p.shape[:2]
        img_size = stride * grid

        p_conf = p[..., 4:5]
        p_cls = p[..., 5:]

        p_d_xywh = p_d[..., :4]

        label_xywh = label[..., :4]
        label_obj_mask = label[..., 4:5]
        label_cls = label[..., 5:]
        #         label_mix = label[..., 5:6]

        # loss ciou, p_d_xywh are the anchors, label_xywh are the labels for the anchors, usually they are [0, 0, 0, 0]
        # if the anchor doesn't have an object associated with it
        ciou = CIOU_xywh_torch(p_d_xywh, label_xywh).unsqueeze(-1)

        # The scaled weight of bbox is used to balance the impact of small objects and large objects on loss.
        bbox_loss_scale = 2.0 - 1.0 * label_xywh[..., 2:3] * label_xywh[..., 3:4] / (
            img_size ** 2
        )
        loss_ciou = label_obj_mask * bbox_loss_scale * (1.0 - ciou)

        # loss confidence
        iou = CIOU_xywh_torch(
            p_d_xywh.unsqueeze(4), bboxes.unsqueeze(1).unsqueeze(1).unsqueeze(1)
        )
        iou_max = iou.max(-1, keepdim=True)[0]
        label_noobj_mask = (1.0 - label_obj_mask) * (
            iou_max < self.__iou_threshold_loss
        ).float()

        loss_conf = label_obj_mask * FOCAL(
            input=p_conf, target=label_obj_mask
        ) + label_noobj_mask * FOCAL(input=p_conf, target=label_obj_mask)

        # loss classes
        loss_cls = label_obj_mask * BCE(input=p_cls, target=label_cls)

        loss_ciou_reduced = (torch.sum(loss_ciou)) / batch_size
        loss_conf_reduced = (torch.sum(loss_conf)) / batch_size
        loss_cls_reduced = (torch.sum(loss_cls)) / batch_size
        loss_reduced = loss_ciou + loss_conf + loss_cls

        if reduce:
            return loss_reduced, loss_ciou_reduced, loss_conf_reduced, loss_cls_reduced
        else:
            # BS x grid_size x grid_size x n_anchors_at_scale x 1 for loss_ciou & loss_conf
            # BS x grid_size x grid_size x n_anchors_at_scale x n_classes for loss_cls
            return loss_ciou, loss_conf, loss_cls


class FocalLoss(nn.Module):
    def __init__(self, gamma=2.0, alpha=1.0, reduction="mean"):
        super(FocalLoss, self).__init__()
        self.__gamma = gamma
        self.__alpha = alpha
        self.__loss = nn.BCEWithLogitsLoss(reduction=reduction)

    def forward(self, input, target):
        loss = self.__loss(input=input, target=target)
        loss *= self.__alpha * torch.pow(
            torch.abs(target - torch.sigmoid(input)), self.__gamma
        )

        return loss


def CIOU_xywh_torch(boxes1, boxes2):
    """
    cal CIOU of two boxes or batch boxes
    :param boxes1:[xmin,ymin,xmax,ymax] or
                [[xmin,ymin,xmax,ymax],[xmin,ymin,xmax,ymax],...]
    :param boxes2:[xmin,ymin,xmax,ymax]
    :return:
    """
    # xywh->xyxy
    boxes1 = torch.cat(
        [
            boxes1[..., :2] - boxes1[..., 2:] * 0.5,
            boxes1[..., :2] + boxes1[..., 2:] * 0.5,
        ],
        dim=-1,
    )
    boxes2 = torch.cat(
        [
            boxes2[..., :2] - boxes2[..., 2:] * 0.5,
            boxes2[..., :2] + boxes2[..., 2:] * 0.5,
        ],
        dim=-1,
    )

    boxes1 = torch.cat(
        [
            torch.min(boxes1[..., :2], boxes1[..., 2:]),
            torch.max(boxes1[..., :2], boxes1[..., 2:]),
        ],
        dim=-1,
    )
    boxes2 = torch.cat(
        [
            torch.min(boxes2[..., :2], boxes2[..., 2:]),
            torch.max(boxes2[..., :2], boxes2[..., 2:]),
        ],
        dim=-1,
    )

    boxes1_area = (boxes1[..., 2] - boxes1[..., 0]) * (boxes1[..., 3] - boxes1[..., 1])
    boxes2_area = (boxes2[..., 2] - boxes2[..., 0]) * (boxes2[..., 3] - boxes2[..., 1])

    inter_left_up = torch.max(boxes1[..., :2], boxes2[..., :2])
    inter_right_down = torch.min(boxes1[..., 2:], boxes2[..., 2:])
    inter_section = torch.max(
        inter_right_down - inter_left_up, torch.zeros_like(inter_right_down)
    )
    inter_area = inter_section[..., 0] * inter_section[..., 1]
    union_area = boxes1_area + boxes2_area - inter_area
    ious = 1.0 * inter_area / union_area

    # cal outer boxes
    outer_left_up = torch.min(boxes1[..., :2], boxes2[..., :2])
    outer_right_down = torch.max(boxes1[..., 2:], boxes2[..., 2:])
    outer = torch.max(
        outer_right_down - outer_left_up, torch.zeros_like(inter_right_down)
    )
    outer_diagonal_line = torch.pow(outer[..., 0], 2) + torch.pow(outer[..., 1], 2)

    # cal center distance
    boxes1_center = (boxes1[..., :2] + boxes1[..., 2:]) * 0.5
    boxes2_center = (boxes2[..., :2] + boxes2[..., 2:]) * 0.5
    center_dis = torch.pow(
        boxes1_center[..., 0] - boxes2_center[..., 0], 2
    ) + torch.pow(boxes1_center[..., 1] - boxes2_center[..., 1], 2)

    # cal penalty term
    # cal width,height
    boxes1_size = torch.max(
        boxes1[..., 2:] - boxes1[..., :2], torch.zeros_like(inter_right_down)
    )
    boxes2_size = torch.max(
        boxes2[..., 2:] - boxes2[..., :2], torch.zeros_like(inter_right_down)
    )
    v = (4 / (math.pi ** 2)) * torch.pow(
        torch.atan((boxes1_size[..., 0] / torch.clamp(boxes1_size[..., 1], min=1e-6)))
        - torch.atan(
            (boxes2_size[..., 0] / torch.clamp(boxes2_size[..., 1], min=1e-6))
        ),
        2,
    )
    alpha = v / (1 - ious + v)

    # cal ciou
    cious = ious - (center_dis / outer_diagonal_line + alpha * v)

    return cious



try:
    from .activeloop import *
    from .activeloop import lookup as activeloop_lookup
    from .from_paths import *
    from .from_paths import lookup as from_paths_lookup

    lookup = {**activeloop_lookup, **from_paths_lookup}
except Exception:
    from .from_paths import *
    from .from_paths import lookup as from_paths_lookup

    lookup = {**from_paths_lookup}



import re
import subprocess
import sys
from functools import reduce
from pathlib import Path

import numpy as np

from ..logger import logger


def compose(functions):
    def wrapper(arg):
        for fn in functions:
            arg = fn(arg)
        return arg

    return wrapper


def is_listy(x):
    return isinstance(x, (tuple, list, np.ndarray))


def if_none(a, b):
    return b if a is None else a


def camel2snake(name):
    _camel_re1 = re.compile("(.)([A-Z][a-z]+)")
    _camel_re2 = re.compile("([a-z0-9])([A-Z])")
    s1 = re.sub(_camel_re1, r"\1_\2", name)
    return re.sub(_camel_re2, r"\1_\2", s1).lower()


def debug_log_on_call(fn):
    def wrapped(*args, **kwargs):
        logger.debug(f"Called: {fn.__qualname__}")
        return fn(*args, **kwargs)

    return wrapped


def find_nearest_greater_divisible(a, b):
    if a % b == 0:
        return a
    return a + (b - (a % b))


def to_numpy(tensor):
    if isinstance(tensor, np.ndarray):
        return tensor
    return tensor.cpu().detach().numpy()


def is_notebook():
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False
    except ModuleNotFoundError:
        return False


def git_commit_hash():
    return subprocess.check_output(["git", "rev-parse", "HEAD"]).strip().decode()


def run_shell_command(cmd, silent=False, same_process=False, background_process=False):
    try:
        if background_process:
            cmd += " &"

        if not silent:
            print(cmd)

        outputs = []
        if same_process:
            process = subprocess.Popen(
                ["bash"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
            )
            process.stdin.write(cmd.encode(sys.stdout.encoding))
            output, _ = process.communicate()
            outputs.append(output)
            return outputs
        else:
            process = subprocess.Popen(
                cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            # output, _ = process.communicate()
            # print(output)
            # sys.stdout.write(output.decode(sys.stdout.encoding))
            for line in iter(process.stdout.readline, b""):
                output = line.decode(sys.stdout.encoding)
                outputs.append(output)
                if not silent:
                    sys.stdout.write(output)
            return outputs
    except subprocess.CalledProcessError as e:
        print(e.output.decode())
        raise e


def catch_all_exceptions(fn):
    def wrapped(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            logger.warning(f"Exception: {str(e)} for {fn} ignored")

    return wrapped


def class_path_with_default(keys, c, default=None):
    try:
        return class_path(keys, c)
    except KeyError:
        return default


def class_path(keys, c):
    if not keys:
        raise ValueError("Expected at least one key, got {0}".format(keys))
    current_value = c
    for key in keys:
        current_value = getattr(current_value, key)
    return current_value


def remove_path(path: Path):
    if path.is_file() or path.is_symlink():
        path.unlink()
        return
    for p in path.iterdir():
        remove_path(p)
    path.rmdir()


from functools import wraps
from time import time


def timing(f, label=None):
    @wraps(f)
    def wrapper(*args, **kwargs):
        start = time()
        result = f(*args, **kwargs)
        end = time()
        if label:
            print(label)
        print(f"Elapsed time: {end - start}")
        return result

    return wrapper


def get_open_port():
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def glob_path(path, glob_string, first_n_paths=None, should_sort=True):
    paths = list(Path(path).glob(glob_string))

    if should_sort:
        paths = sorted(paths)

    if first_n_paths is not None:
        paths = paths[0:first_n_paths]

    return paths


lookup = {
    "is_listy": is_listy,
    "if_none": if_none,
    "camel2snake": camel2snake,
    "debug_log_on_call": debug_log_on_call,
    "find_nearest_greater_divisible": find_nearest_greater_divisible,
    "to_numpy": to_numpy,
    "is_notebook": is_notebook,
    "catch_all_exceptions": catch_all_exceptions,
    "remove_path": remove_path,
    "timing": timing,
    "get_open_port": get_open_port,
    "glob_path": glob_path,
    "compose": compose,
}



from .adapters import *
from .adapters import lookup as adapters_lookup
from .callbacks import *
from .callbacks import lookup as callbacks_lookup
from .trainer import *
from .trainer import lookup as trainer_lookup

lookup = {**adapters_lookup, **callbacks_lookup, **trainer_lookup}



from dataclasses import dataclass

import yaml


class RoheboamYamlLoader(yaml.SafeLoader):
    pass


class RoheboamYamlDumper(yaml.SafeDumper):
    pass


@dataclass
class Variable:
    variable_group: str
    variable_name: str

    @property
    def is_nested(self):
        return self.variable_group is not None

    @property
    def variable_groups(self):
        return [subgroup_name for subgroup_name in self.variable_group.split(".")]

    def __hash__(self):
        return hash(f"{self.variable_name}{self.variable_group}")

    def __eq__(self, other):
        if not isinstance(self, type(other)):
            return False

        if self.variable_group != other.variable_group:
            return False

        return self.variable_name == other.variable_name


def var_constructor(loader, node):
    if "." in node.value:
        *variable_group, variable_name = node.value.split(".")
        return Variable(
            variable_group=".".join(variable_group), variable_name=variable_name
        )
    else:
        return Variable(variable_group=None, variable_name=node.value)


def var_representer(dumper: RoheboamYamlDumper, variable: Variable):
    if variable.is_nested:
        value = f"{variable.variable_group}.{variable.variable_name}"
    else:
        value = f"{variable.variable_name}"
    return dumper.represent_scalar("!Var", f"{value}")


@dataclass
class Reference:
    ref_node_name: str
    output_name: str


def ref_constructor(loader, node):
    ref_node_name_and_output_name = node.value.split(".")
    assert (
        len(ref_node_name_and_output_name) == 2
    ), f"No output is being accessed for resource: '{node.tag} {node.value}' please search \
        through the configuration file using '{node.tag} {node.value}' to add an access to that resource"
    ref_node_name, output_name = ref_node_name_and_output_name
    return Reference(ref_node_name=ref_node_name, output_name=output_name)


def ref_representer(dumper, data):
    value = f"{data.ref_node_name}.{data.output_name}"
    return dumper.represent_scalar("!Ref", f"{value}")


RoheboamYamlLoader.add_constructor("!Var", var_constructor)
RoheboamYamlLoader.add_constructor("!Ref", ref_constructor)

RoheboamYamlDumper.add_representer(Variable, var_representer)
RoheboamYamlDumper.add_representer(Reference, ref_representer)



# Ranger deep learning optimizer - RAdam + Lookahead + Gradient Centralization, combined into one optimizer.

# https://github.com/lessw2020/Ranger-Deep-Learning-Optimizer
# and/or
# https://github.com/lessw2020/Best-Deep-Learning-Optimizers

# Ranger has now been used to capture 12 records on the FastAI leaderboard.

# This version = 20.4.11

# Credits:
# Gradient Centralization --> https://arxiv.org/abs/2004.01461v2 (a new optimization technique for DNNs),
# github:  https://github.com/Yonghongwei/Gradient-Centralization
# RAdam -->  https://github.com/LiyuanLucasLiu/RAdam
# Lookahead --> rewritten by lessw2020, but big thanks to Github @LonePatient and @RWightman for ideas from their code.
# Lookahead paper --> MZhang,G Hinton  https://arxiv.org/abs/1907.08610

# summary of changes:
# 4/11/20 - add gradient centralization option.  Set new testing benchmark for accuracy with it, toggle with use_gc flag at init.
# full code integration with all updates at param level instead of group, moves slow weights into state dict (from generic weights),
# supports group learning rates (thanks @SHolderbach), fixes sporadic load from saved model issues.
# changes 8/31/19 - fix references to *self*.N_sma_threshold;
# changed eps to 1e-5 as better default than 1e-8.

import math

import torch
from torch.optim.optimizer import Optimizer, required


class Ranger(Optimizer):
    def __init__(
        self,
        params,
        lr=1e-3,  # lr
        alpha=0.5,
        k=6,
        N_sma_threshhold=5,  # Ranger options
        betas=(0.95, 0.999),
        eps=1e-5,
        weight_decay=0,  # Adam options
        # Gradient centralization on or off, applied to conv layers only or conv + fc layers
        use_gc=True,
        gc_conv_only=False,
    ):

        # parameter checks
        if not 0.0 <= alpha <= 1.0:
            raise ValueError(f"Invalid slow update rate: {alpha}")
        if not 1 <= k:
            raise ValueError(f"Invalid lookahead steps: {k}")
        if not lr > 0:
            raise ValueError(f"Invalid Learning Rate: {lr}")
        if not eps > 0:
            raise ValueError(f"Invalid eps: {eps}")

        # parameter comments:
        # beta1 (momentum) of .95 seems to work better than .90...
        # N_sma_threshold of 5 seems better in testing than 4.
        # In both cases, worth testing on your dataset (.90 vs .95, 4 vs 5) to make sure which works best for you.

        # prep defaults and init torch.optim base
        defaults = dict(
            lr=lr,
            alpha=alpha,
            k=k,
            step_counter=0,
            betas=betas,
            N_sma_threshhold=N_sma_threshhold,
            eps=eps,
            weight_decay=weight_decay,
        )
        super().__init__(params, defaults)

        # adjustable threshold
        self.N_sma_threshhold = N_sma_threshhold

        # look ahead params

        self.alpha = alpha
        self.k = k

        # radam buffer for state
        self.radam_buffer = [[None, None, None] for ind in range(10)]

        # gc on or off
        self.use_gc = use_gc

        # level of gradient centralization
        self.gc_gradient_threshold = 3 if gc_conv_only else 1

        print(
            f"Ranger optimizer loaded. \nGradient Centralization usage = {self.use_gc}"
        )
        if self.use_gc and self.gc_gradient_threshold == 1:
            print("GC applied to both conv and fc layers")
        elif self.use_gc and self.gc_gradient_threshold == 3:
            print("GC applied to conv layers only")

    def __setstate__(self, state):
        print("set state called")
        super(Ranger, self).__setstate__(state)

    def step(self, closure=None):
        loss = None

        if closure is not None:
            loss = closure()

        # Evaluate averages and grad, update param tensors
        for group in self.param_groups:

            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()

                if grad.is_sparse:
                    raise RuntimeError(
                        "Ranger optimizer does not support sparse gradients"
                    )

                p_data_fp32 = p.data.float()

                state = self.state[p]  # get state dict for this param

                if (
                    len(state) == 0
                ):  # if first time to run...init dictionary with our desired entries
                    # if self.first_run_check==0:
                    # self.first_run_check=1
                    # print("Initializing slow buffer...should not see this at load from saved model!")
                    state["step"] = 0
                    state["exp_avg"] = torch.zeros_like(p_data_fp32)
                    state["exp_avg_sq"] = torch.zeros_like(p_data_fp32)

                    # look ahead weight storage now in state dict
                    state["slow_buffer"] = torch.empty_like(p.data)
                    state["slow_buffer"].copy_(p.data)

                else:
                    state["exp_avg"] = state["exp_avg"].type_as(p_data_fp32)
                    state["exp_avg_sq"] = state["exp_avg_sq"].type_as(p_data_fp32)

                # begin computations
                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                # GC operation for Conv layers and FC layers
                if grad.dim() > self.gc_gradient_threshold:
                    grad.add_(-grad.mean(dim=tuple(range(1, grad.dim())), keepdim=True))

                state["step"] += 1

                # compute variance mov avg
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                # compute mean moving avg
                exp_avg.mul_(beta1).add_(1 - beta1, grad)

                buffered = self.radam_buffer[int(state["step"] % 10)]

                if state["step"] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state["step"]
                    beta2_t = beta2 ** state["step"]
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * state["step"] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma
                    if N_sma > self.N_sma_threshhold:
                        step_size = math.sqrt(
                            (1 - beta2_t)
                            * (N_sma - 4)
                            / (N_sma_max - 4)
                            * (N_sma - 2)
                            / N_sma
                            * N_sma_max
                            / (N_sma_max - 2)
                        ) / (1 - beta1 ** state["step"])
                    else:
                        step_size = 1.0 / (1 - beta1 ** state["step"])
                    buffered[2] = step_size

                if group["weight_decay"] != 0:
                    p_data_fp32.add_(-group["weight_decay"] * group["lr"], p_data_fp32)

                # apply lr
                if N_sma > self.N_sma_threshhold:
                    denom = exp_avg_sq.sqrt().add_(group["eps"])
                    p_data_fp32.addcdiv_(-step_size * group["lr"], exp_avg, denom)
                else:
                    p_data_fp32.add_(-step_size * group["lr"], exp_avg)

                p.data.copy_(p_data_fp32)

                # integrated look ahead...
                # we do it at the param level instead of group level
                if state["step"] % group["k"] == 0:
                    # get access to slow param tensor
                    slow_p = state["slow_buffer"]
                    # (fast weights - slow weights) * alpha
                    slow_p.add_(self.alpha, p.data - slow_p)
                    # copy interpolated weights to RAdam param tensor
                    p.data.copy_(slow_p)

        return loss



import pdb
import platform
import sys

from setuptools import find_packages, setup

import roheboam

setup(
    name="roheboam",
    version=roheboam.__version__,
    author="Kevin Lu",
    author_email="kevinyihchyunlu@gmail.com",
    license="GPL3",
    packages=find_packages(exclude=["tests", "tests/*", "examples", "examples/*"]),
    entry_points={
        "console_scripts": [
            "roheboam = roheboam.cli:cli",
        ],
    },
    install_requires=[],
    zip_safe=False,
)



from abc import ABC
from copy import deepcopy
from functools import partial

import numpy as np

from ...utils.convenience import is_listy


class Data(ABC):
    def __init__(
        self,
        data=None,
        path=None,
        load_fn=None,
        name=None,
        should_cache=False,
        fns_to_run=None,
    ):
        if data is None and path is None:
            assert (
                False
            ), "Both the data and the path to the data can't be none, you must define one"

        if path is not None:
            assert (
                load_fn is not None
            ), "A path is provided, so a function must be defined to load it"

        self._data = data
        self._original_data = deepcopy(data)
        self.path = path
        self.load_fn = load_fn
        self.name = name
        self.should_cache = should_cache
        if fns_to_run is None:
            self.fns_to_run = []

    @property
    def original_data(self):
        return self._original_data

    @property
    def data(self):
        return self._load_if_needed_then_cache()

    # def run_fns_to_run(self):
    #     while len(self.fns_to_run) > 0:
    #         self.fns_to_run.pop(0)()

    @data.setter
    def data(self, value):
        self._data = value

    # # TODO: fix this, it shouldn't be here since not all labels are lists
    # def filter_by_idx(self, idx, lazy=True):
    #     if len(idx) == 0:
    #         self._data = np.array([])
    #         return

    #     if not lazy:
    #         self._data = self.data[idx]
    #         return

    #     self.fns_to_run.append(partial(self.filter_by_idx, idx=idx, lazy=False))

    # def transform_labels(self, transform_label_map, lazy=True):
    #     if not lazy:
    #         print(self.__dict__)
    #         self._data = np.array([transform_label_map[d] for d in self.data])
    #         return

    #     self.fns_to_run.append(partial(self.transform_labels, transform_label_map=transform_label_map, lazy=False))

    def _load_if_needed_then_cache(self):
        if self._data is None:
            if is_listy(self.path):
                data = [self.load_fn(p) for p in self.path]
            else:
                data = self.load_fn(self.path)
            if self.should_cache:
                self._original_data = deepcopy(data)
                self._data = data
            return data
        else:
            return self._data


lookup = {"Data": Data}



from collections import defaultdict

from ....utils.convenience import to_numpy
from ..callback import Callback
from .utils import ShouldRecordMixin


class ImageSegmentationResultRecorder(Callback, ShouldRecordMixin):

    _order = -10

    def __init__(
        self,
        train_merge_tiles_fn=lambda x: x,
        inference_merge_tiles_fn=lambda x: x,
        convert_tensor_to_image_fn=lambda x: x,
        convert_tensor_to_mask_fn=lambda x: x,
        record_on_validation=True,
        record_on_test=True,
        record_every_n_steps=25,
        name=None,
    ):
        super().__init__(name)
        self.train_merge_tiles_fn = train_merge_tiles_fn
        self.inference_merge_tiles_fn = inference_merge_tiles_fn
        self.convert_tensor_to_image_fn = convert_tensor_to_image_fn
        self.convert_tensor_to_mask_fn = convert_tensor_to_mask_fn
        self.record_on_validation = record_on_validation
        self.record_on_test = record_on_test
        self.record_every_n_steps = record_every_n_steps
        self.results = defaultdict(list)

    def on_train_batch_end(self, context):

        phase, epoch, step, batch, predictions = (
            context.current_phase,
            context.current_epoch,
            context.current_step,
            context.train_batch,
            context.train_prediction,
        )

        if not self._should_store(phase, epoch, step):
            return

        images = self.train_merge_tiles_fn(
            self.convert_tensor_to_image_fn(batch["images"])
        )
        raw_predictions = self.train_merge_tiles_fn(
            to_numpy(predictions).transpose(0, 2, 3, 1)
        )
        final_predictions = self.train_merge_tiles_fn(
            self.convert_tensor_to_mask_fn(predictions)
        )
        masks = self.train_merge_tiles_fn(
            self.convert_tensor_to_mask_fn(batch["masks"])
        )
        self.results[(str(phase), epoch)].extend(
            [
                {
                    "image": image,
                    "prediction": raw_prediction,
                    "final_prediction": final_prediction,
                    "mask": mask,
                }
                for image, raw_prediction, final_prediction, mask in zip(
                    images, raw_predictions, final_predictions, masks
                )
            ]
        )

    def on_validation_batch_end(self, context):
        phase, epoch, step, batch, predictions = (
            context.current_phase,
            context.current_epoch,
            context.current_step,
            context.validation_batch,
            context.validation_prediction,
        )
        if not self._should_store(phase, epoch, step):
            return

        images = self.inference_merge_tiles_fn(
            self.convert_tensor_to_image_fn(batch["images"])
        )
        raw_predictions = self.inference_merge_tiles_fn(
            to_numpy(predictions).transpose(0, 2, 3, 1)
        )
        final_predictions = self.inference_merge_tiles_fn(
            self.convert_tensor_to_mask_fn(predictions)
        )
        masks = self.inference_merge_tiles_fn(
            self.convert_tensor_to_mask_fn(batch["masks"])
        )
        self.results[str(phase), epoch].extend(
            [
                {
                    "image": image,
                    "prediction": raw_prediction,
                    "final_prediction": final_prediction,
                    "mask": mask,
                }
                for image, raw_prediction, final_prediction, mask in zip(
                    images, raw_predictions, final_predictions, masks
                )
            ]
        )

    def on_test_batch_end(self, context):
        phase, epoch, step, batch, predictions = (
            context.current_phase,
            context.current_epoch,
            context.current_step,
            context.test_batch,
            context.test_prediction,
        )
        if not self._should_store(phase, epoch, step):
            return

        images = self.inference_merge_tiles_fn(
            self.convert_tensor_to_image_fn(batch["images"])
        )
        raw_predictions = self.inference_merge_tiles_fn(
            to_numpy(predictions).transpose(0, 2, 3, 1)
        )
        final_predictions = self.inference_merge_tiles_fn(
            self.convert_tensor_to_mask_fn(predictions)
        )
        has_masks = "masks" in batch

        if has_masks:
            masks = self.inference_merge_tiles_fn(
                self.convert_tensor_to_mask_fn(batch["masks"])
            )
            result = [
                {
                    "image": image,
                    "prediction": raw_prediction,
                    "final_prediction": final_prediction,
                    "mask": mask,
                }
                for image, raw_prediction, final_prediction, mask in zip(
                    images, raw_predictions, final_predictions, masks
                )
            ]
        else:
            result = [
                {
                    "image": image,
                    "prediction": raw_prediction,
                    "final_prediction": final_prediction,
                }
                for image, raw_prediction, final_prediction, in zip(
                    images, raw_predictions, final_predictions
                )
            ]
        self.results[(str(phase), epoch)].extend(result)


class ImageSegmentationInferenceRecorder(Callback):

    _order = -10

    def __init__(
        self,
        inference_merge_tiles_fn=lambda x: x,
        convert_tensor_to_image_fn=lambda x: x,
        convert_tensor_to_mask_fn=lambda x: x,
        name=None,
    ):
        super().__init__(name)
        self.inference_merge_tiles_fn = inference_merge_tiles_fn
        self.convert_tensor_to_image_fn = convert_tensor_to_image_fn
        self.convert_tensor_to_mask_fn = convert_tensor_to_mask_fn
        self.results = []

    @property
    def output(self):
        return dict(
            names=self.names,
            input=self.inputs,
            predicted_prob_for_masks=self.predicted_prob_for_masks,
            predicted_masks=self.predicted_masks,
            true_masks=self.true_masks,
        )

    @property
    def serializable_results(self):
        serializable_results = []
        for result in self.results:
            serializable_results.append({k: v.tolist() for k, v in result.items()})
        return serializable_results

    def flush_results(self):
        self.results = []

    def on_test_batch_end(self, context):
        _, _, batch, predictions = (
            context.current_phase,
            context.current_epoch,
            context.test_batch,
            context.test_prediction,
        )
        images = self.convert_tensor_to_image_fn(batch["images"])
        raw_predictions = self.inference_merge_tiles_fn(to_numpy(predictions))
        final_predictions = self.inference_merge_tiles_fn(
            self.convert_tensor_to_mask_fn(predictions)
        )
        self.results.extend(
            [
                {
                    "image": image,
                    "prediction": raw_prediction,
                    "final_prediction": final_prediction,
                }
                for image, raw_prediction, final_prediction in zip(
                    images, raw_predictions, final_predictions
                )
            ]
        )



from pathlib import Path
from uuid import uuid4

import matplotlib.pyplot as plt
import numpy as np

from ......core.data import Data
from ......vision.utils import (
    clip_bboxes,
    cxcywh_to_xyxy,
    denormalize_bboxes_for_image,
    imread_rgb,
    plot_bounding_boxes_on_image_with_label,
    xywh_to_xyxy,
)


class _DataMixin:
    @property
    def data(self):
        return (
            self.image.data,
            self.bboxes.data if self.bboxes_exists else None,
            self.labels.data if self.labels_exists else None,
            self.name if self.name is not None else None,
        )

    @data.setter
    def data(self, image=None, bboxes=None, labels=None):
        if image is not None:
            self.image.data = image
        if bboxes is not None:
            self.bboxes.data = bboxes
        if labels is not None:
            self.labels.data = labels


class _PlotMixin:
    def plot(self, labels_to_plot=None, figsize=(8, 6)):
        if self.bboxes_exists and self.labels_exists:
            if labels_to_plot is not None:
                idxs_to_keep = [
                    i for (i, l) in enumerate(self.labels.data) if l in labels_to_plot
                ]
                plt.figure(figsize=figsize)
                plt.imshow(
                    plot_bounding_boxes_on_image_with_label(
                        self.image.data,
                        self.pascal_voc_bboxes[idxs_to_keep],
                        self.labels.data[idxs_to_keep],
                    )
                )
            else:
                plt.figure(figsize=figsize)
                plt.imshow(
                    plot_bounding_boxes_on_image_with_label(
                        self.image.data, self.pascal_voc_bboxes, self.labels.data
                    )
                )
        else:
            plt.imshow(self.image.data)

    def create_plot_array(self, labels_to_plot=None):
        if self.bboxes_exists and self.labels_exists:
            if labels_to_plot is not None:
                idxs_to_keep = [
                    i for (i, l) in enumerate(self.labels.data) if l in labels_to_plot
                ]
                return plot_bounding_boxes_on_image_with_label(
                    self.image.data,
                    self.pascal_voc_bboxes[idxs_to_keep],
                    self.labels.data[idxs_to_keep],
                )
            else:
                return plot_bounding_boxes_on_image_with_label(
                    self.image.data, self.pascal_voc_bboxes, self.labels.data
                )
        else:
            return self.image.data


class _BoundingBoxMixin:
    @property
    def labels(self):
        # self._labels.run_fns_to_run()
        return self._labels

    @property
    def bboxes_format(self):
        return self._bboxes_format

    @property
    def bboxes(self):
        if self._convert_to_pascal_voc_bboxes_in_future:
            self.convert_to_pascal_voc_bboxes(lazy=False)
            self._convert_to_pascal_voc_bboxes_in_future = False
            self._bboxes_format = "PASCAL_VOC"
        # self._bboxes.run_fns_to_run()
        return self._bboxes

    @property
    def pascal_voc_bboxes(self):
        height, width, _ = self.image.data.shape
        bboxes_format = self.bboxes_format
        bboxes = self.bboxes.data
        return self._create_pascal_voc_bboxes(height, width, bboxes, bboxes_format)

    def convert_to_pascal_voc_bboxes(self, lazy=True):
        if self.bboxes_exists and lazy:
            self._convert_to_pascal_voc_bboxes_in_future = True

        if self.bboxes_exists and not lazy:
            height, width, _ = self.image.data.shape
            self._bboxes.data = self._create_pascal_voc_bboxes(
                height, width, self._bboxes.data, self._bboxes_format
            )
            self._bboxes_format = "PASCAL_VOC"

    def _create_pascal_voc_bboxes(
        self, image_height, image_width, bboxes, bboxes_format
    ):
        if bboxes_format == "PASCAL_VOC":
            return bboxes
        elif bboxes_format == "YOLO":
            bboxes = cxcywh_to_xyxy(
                denormalize_bboxes_for_image(
                    bboxes, height=image_height, width=image_width
                )
            )
        elif bboxes_format == "COCO":
            bboxes = xywh_to_xyxy(bboxes)

        if len(bboxes) > 0:
            bboxes = clip_bboxes(bboxes, image_height, image_width)

        return bboxes

    @property
    def has_labels(self):
        return self.labels_exists and len(self.labels.data) > 0

    @property
    def labels_exists(self):
        return self._labels is not None

    @property
    def bboxes_area(self):
        bboxes = self.pascal_voc_bboxes
        return (bboxes[:, 3] - bboxes[:, 1]) * (bboxes[:, 2] - bboxes[:, 0])

    @property
    def has_bboxes(self):
        return self.bboxes_exists and len(self.bboxes.data) > 0

    @property
    def bboxes_exists(self):
        return self._bboxes is not None

    def idx_of_unique_bboxes(self):
        unique_idxs = []
        bboxes_id = [
            f"{bbox.tolist()}{label}"
            for (bbox, label) in zip(self.bboxes.data, self.labels.data)
        ]
        bboxes_map = {}
        for i, bbox_id in enumerate(bboxes_id):
            if bbox_id not in bboxes_map:
                bboxes_map[bbox_id] = True
                unique_idxs.append(i)
        return np.array(unique_idxs)

    def idx_of_bboxes_below_pixel_area(self, area=8):
        return np.where(self.bboxes_area < area)[0]

    def idx_of_bboxes_above_or_equal_pixel_area(self, area=8):
        return np.where(self.bboxes_area >= area)[0]

    def remove_duplicate_bboxes(self):
        idxs_to_keep = self.idx_of_unique_bboxes()
        if self.labels is not None and self.bboxes is not None:
            self._labels.data = self._labels.data[idxs_to_keep]
            self._bboxes.data = self._bboxes.data[idxs_to_keep]
        return self

    def remove_bboxes_below_pixel_area(self, area=8):
        idxs_to_keep = self.idx_of_bboxes_above_or_equal_pixel_area(area)
        if self.labels is not None and self.bboxes is not None:
            self._labels.data = self._labels.data[idxs_to_keep]
            self._bboxes.data = self._bboxes.data[idxs_to_keep]
        return self

    def filter_labels_and_bboxes_to_keep(self, labels_to_keep, lazy=True):
        idxs_to_keep = []
        for i, label in enumerate(self._labels.data):
            if label in labels_to_keep:
                idxs_to_keep.append(i)

        # self._labels.filter_by_idx(idx=idxs_to_keep, lazy=lazy)
        # self._bboxes.filter_by_idx(idx=idxs_to_keep, lazy=lazy)

        self = filter_bboxes_and_labels_by_idx_for_sample(self, idx=idxs_to_keep)

    def transform_labels(self, transform_label_map, lazy=True):
        # self._labels.transform_labels(transform_label_map, lazy=lazy)
        self = transform_bboxes_labels_for_sample(self, transform_label_map)


def filter_labels_and_bboxes_to_keep_for_sample(sample, labels_to_keep):
    idxs_to_keep = []
    for i, label in enumerate(sample.labels.data):
        if label in labels_to_keep:
            idxs_to_keep.append(i)

    sample = filter_bboxes_and_labels_by_idx_for_sample(sample, idx=idxs_to_keep)
    return sample


def filter_bboxes_and_labels_by_idx_for_sample(sample, idx):
    if len(idx) == 0:
        return np.array([])

    sample.bboxes.data = sample.bboxes.data[idx]
    sample.labels.data = sample.labels.data[idx]
    return sample


def transform_bboxes_labels_for_sample(sample, transform_label_map):
    sample.labels.data = np.array([transform_label_map[d] for d in sample.labels.data])
    return sample


class ImageBoundingBoxDetectionSample(_DataMixin, _PlotMixin, _BoundingBoxMixin):
    """
    PASCAL_VOC: [x_min, y_min, x_max, y_max]

    COCO: [x_min, y_min, width, height]

    YOLO: normalized[center_x, center_y, width, height]
    """

    def __init__(self, image, labels=None, bboxes=None, bboxes_format=None, name=None):
        if labels is not None and bboxes is not None:
            assert len(bboxes.data) == len(
                labels.data
            ), f"The number of bboxes ({len(bboxes.data)}) must be the same as the number of labels ({len(labels.data)})"

        assert not (
            (bboxes is None) ^ (bboxes_format is None)
        ), "Both bboxes and bboxes_format must be defined together, or not defined at all"

        assert bboxes_format in [
            "PASCAL_VOC",
            "COCO",
            "YOLO",
            None,
        ], "bboxes_format needs to be 'PASCAL_VOC', 'COCO', or 'YOLO'"

        self.image = image
        self._labels = labels
        self._bboxes = bboxes
        self._bboxes_format = bboxes_format
        self._convert_to_pascal_voc_bboxes_in_future = False
        self._filter_labels_and_bboxes_to_keep_in_future = False
        self._filter_labels_and_bboxes_to_keep_in_future_args = {}
        self.name = str(uuid4()) if name is None else name

    @classmethod
    def create(
        cls,
        image_data=None,
        image_path=None,
        load_image_fn=imread_rgb,
        labels_data=None,
        labels_path=None,
        load_labels_fn=None,
        bboxes_data=None,
        bboxes_path=None,
        load_bboxes_fn=None,
        bboxes_format=None,
        name=None,
    ):
        if name is None:
            name = Path(image_path).stem if image_path is not None else str(uuid4())

        if image_data is not None:
            image_data = np.array(image_data)

        image = Data(image_data, image_path, load_image_fn)

        if labels_data is None and labels_path is None:
            labels = None
        else:
            if labels_data is not None:
                labels_data = np.array(labels_data)

            labels = Data(labels_data, labels_path, load_labels_fn)

        if bboxes_data is None and bboxes_path is None:
            bboxes = None
        else:
            if bboxes_data is not None:
                bboxes_data = np.array(bboxes_data)
            bboxes = Data(bboxes_data, bboxes_path, load_bboxes_fn)
        return cls(image, labels, bboxes, bboxes_format, name)


lookup = {
    "ImageBoundingBoxDetectionSample": ImageBoundingBoxDetectionSample,
    "transform_bboxes_labels_for_sample": transform_bboxes_labels_for_sample,
    "filter_bboxes_and_labels_by_idx_for_sample": filter_bboxes_and_labels_by_idx_for_sample,
    "filter_labels_and_bboxes_to_keep_for_sample": filter_labels_and_bboxes_to_keep_for_sample,
}



import time

import requests

from ..engine.utils.convenience import run_shell_command


def post_request_on_url_with_retry_until_success(
    url, data, attempts=10, per_try_timeout=5, completion_handler=lambda: None
):
    completed_successfully = False
    try:
        for attempt in range(attempts):
            try:
                response = requests.post(
                    url,
                    data=data,
                    headers={"Content-Type": "application/json"},
                )
                if response.status_code != 200:
                    raise Exception(
                        f"Response status code: {response.status_code}\nResponse text: {response.text}"
                    )
                assert response.status_code == 200
                print(f"Success on attempt: {attempt + 1}")
                completed_successfully = True
                break
            except Exception as e:
                print(f"Failure on attempt: {attempt + 1}")
                print(e)
                time.sleep(per_try_timeout)
    finally:
        completion_handler()
        return completed_successfully


def stop_container_completion_handler(container, silent=True):
    try:
        if not silent:
            print(container.logs().decode())
        container.stop()
    except Exception as e:
        print(e)


def stop_k8s_pods_completion_handler(
    deployment_manifest_save_path, gateway_manifest_save_path
):
    try:
        run_shell_command(f"microk8s kubectl delete -f {deployment_manifest_save_path}")
        run_shell_command(f"microk8s kubectl delete -f {gateway_manifest_save_path}")
    except Exception as e:
        print(e)



from torch.nn import functional as F

from . import BaseLoss


class FocalLoss(BaseLoss):
    def __init__(
        self,
        name=None,
        gamma=2,
        multilabel=False,
        per_sample_loss_aggregate_method="SUM",
    ):
        super().__init__(name)
        self.gamma = gamma
        self.multilabel = multilabel
        self.per_sample_loss_aggregate_method = per_sample_loss_aggregate_method

    # def to_call_kwargs(self, d):
    #     predictions = d["predictions"]
    #     targets = d["targets"]
    #     auxillary = d["auxillary"]

    #     method_args = {"predictions": d["predictions"]}

    #     if auxillary is not None and "weight_maps" in d["auxillary"]:
    #         method_args["weight_maps"] = d["auxillary"]["weight_maps"]["data"]

    #     if "masks" in d["targets"]:
    #         method_args["targets"] = d["targets"]["masks"]["data"]
    #         return method_args

    #     if "labels" in d["targets"]:
    #         method_args["targets"] = d["targets"]["labels"]["data"]
    #         return method_args

    #     raise MissingMethodArgumentException(
    #         f"There must be a 'd['targets']['labels'] or 'd['targets']['masks']
    #            key defined in the keys\n: {d['targets'].keys()}\nprovided by the dictionary"
    #     )

    @property
    def unreduced_loss(self):
        return self._unreduced_loss

    @property
    def per_sample_loss(self):
        return self._per_sample_loss

    @property
    def reduced_loss(self):
        return self._reduced_loss

    @property
    def total_loss(self):
        return self._total_loss

    def __call__(self, predictions, targets, auxillary={}):
        # This returns B x ... (same shape as input)
        weight_maps = auxillary.get("weight_maps")
        if weight_maps is not None:
            self._unreduced_loss = weight_maps * focal_loss(
                predictions, targets, self.gamma
            )
        else:
            self._unreduced_loss = focal_loss(predictions, targets, self.gamma)

        self._per_sample_loss = (
            self.reshape_to_batch_size_x_minus_one_aggregate_over_last_dimension(
                self._unreduced_loss,
                aggregate_method=self.per_sample_loss_aggregate_method,
            )
        )
        self._reduced_loss = self._per_sample_loss.mean()
        self._total_loss = self._reduced_loss
        return self._total_loss


class CrossEntropyLoss(FocalLoss):
    def __init__(
        self, name=None, multilabel=False, per_sample_loss_aggregate_method="SUM"
    ):
        super().__init__(
            name,
            gamma=0,
            multilabel=multilabel,
            per_sample_loss_aggregate_method=per_sample_loss_aggregate_method,
        )


def focal_loss(input, target, gamma=2, multilabel=False):
    """

    :param input: B x N_classes x ...
    :param target: B x N_classes x ... # one hot encoding
    :param gamma: the higher the value, the greater the loss for uncertain classes
    :return:
    """

    if not (target.size() == input.size()):
        target = F.one_hot(target, num_classes=input.size(-1))

    target = target.float()

    if multilabel:
        max_val = (-input).clamp(min=0)
        loss = (
            input
            - input * target
            + max_val
            + ((-max_val).exp() + (-input - max_val).exp()).log()
        )
        inv_probs = F.logsigmoid(-input * (target * 2.0 - 1.0))
        loss = (inv_probs * gamma).exp() * loss
    else:
        loss = -1.0 * F.log_softmax(input, dim=-1) * target
    return loss
